#!/usr/bin/env python
#
import os
import femagtools.erg
import numpy as np


def read_erg(filename):
    testPath = os.path.join(os.path.split(__file__)[0], 'data')
    if len(testPath) == 0:
        testPath = os.path.join(os.path.abspath('.'), 'data')
    r = femagtools.erg.read(os.path.join(testPath, filename))
    return r


def test_read_erg():
    r = read_erg('ldlq.erg')
    assert len(r.keys()) == 14
    assert min(r['beta']) == -90.0
    assert max(r['beta']) == 0.0
    np.array(r['M_FE']).shape == (10, 10)




#!/usr/bin/env python
#
import femagtools.airgap as ag
import numpy as np
import os


def test_airgap_induction():
    testPath = os.path.join(os.path.split(__file__)[0], 'data')
    if len(testPath) == 0:
        testPath = os.path.join(os.path.abspath('.'), 'data')
    r = ag.read(os.path.join(testPath, 'bag.dat'))
    np.testing.assert_almost_equal(r['Bamp'], 1.26914, 3)
    assert r['npoles'] == 8
    harms = [nue for nue, amp in zip(r['nue'],
                                     r['B_nue']) if amp > 1e-2]

    assert harms == [4, 12, 20, 28, 36, 44, 52, 60, 68]

def test_airgap_induction2():
    testPath = os.path.join(os.path.split(__file__)[0], 'data')
    if len(testPath) == 0:
        testPath = os.path.join(os.path.abspath('.'), 'data')
    r = ag.read(os.path.join(testPath, 'bag2.dat'))
    np.testing.assert_almost_equal(r['Bamp'], 0.9272, 3)
    assert r['npoles'] == 32
    harms = [nue for nue, amp in zip(r['nue'],
                                     r['B_nue']) if amp > 1e-2]

    assert harms == [16, 48, 80, 112, 144, 176, 208, 272]



import femagtools.mcv

def test_write_poc(tmpdir):
    filename = f"{tmpdir}/mcvData"

    B = [3e-02, 7e-02, 2e-01, 2.31]
    H = [1e+02, 2e+02, 5e+02, 3e+05]
    mcvData = dict(curve=[dict(bi=B,hi=H)])
    mcv = femagtools.mcv.MagnetizingCurve(mcvData)
    mcv.writefile(filename)



import numpy as np
import scipy.interpolate as ip
import logging
import multiprocessing
from .utils import betai1
from .pm import PmRelMachineLdq, PmRelMachinePsidq, PmRelMachine
from .sm import SynchronousMachine, SynchronousMachineLdq, SynchronousMachinePsidq
from . import create_from_eecpars

logger = logging.getLogger("femagtools.effloss")


def iqd_tmech_umax(m, u1, with_mtpa, progress, speed_torque, iq, id, iex):
    """calculate iq, id for each load (n, T) from speed_torque at voltage u1

        Arguments:
        m: PmRelMachine or SynchronousMachine
        u1: (float) phase voltage (V)
        speed_torque: list of (n, T) pairs
        with_mtpa: (bool) use mtpa in const flux range if True
        progress: logging pipe """
    nsamples = len(speed_torque)
    num_iv = round(nsamples/7)
    try:
        for i, nT in enumerate(speed_torque):
            iqde = m.iqd_tmech_umax(
                nT[1], 2*np.pi*nT[0]*m.p,
                u1, with_mtpa=with_mtpa)[:-1]
            iq[i] = iqde[0]
            id[i] = iqde[1]
            if len(iqde) > 2:
                iex[i] = iqde[2]
            if i % num_iv == 0:
                progress.send(f"{100*i/nsamples:.1f}%")
    except Exception as e:
        progress.send(e)
    finally:
        progress.close()

def iqd_tmech_umax_multi(num_proc, ntmesh, m, u1, with_mtpa):
    """calculate iqd for sm and pm using multiproc
    """
    progress_readers = []
    chunksize = int(np.ceil(ntmesh.shape[1]/num_proc))
    procs = []
    iq = []
    id = []
    iex = []
    iexk = []
    k = 0
    for i in range(0, num_proc*chunksize, chunksize):
        prog_reader, prog_writer = multiprocessing.Pipe(duplex=False)
        progress_readers.append(prog_reader)
        iq.append(multiprocessing.Array('d', chunksize))
        id.append(multiprocessing.Array('d', chunksize))
        if isinstance(m, SynchronousMachine):
            iex.append(multiprocessing.Array('d', chunksize))
            iexk = iex[k]
        p = multiprocessing.Process(target=iqd_tmech_umax,
                                    args=(m, u1, with_mtpa,
                                          prog_writer,
                                          ntmesh.T[i:i+chunksize],
                                          iq[k], id[k], iexk))
        k += 1
        p.start()
        procs.append(p)
        prog_writer.close()

    i = 0
    collected_msg = []
    while progress_readers:
        for r in multiprocessing.connection.wait(progress_readers):
            try:
                collected_msg.append(r.recv())
                i += 1
            except EOFError:
                progress_readers.remove(r)
            else:
                if i % len(progress_readers) == 0:
                    logger.info("Losses/Eff Map: %s",
                                ', '.join(collected_msg))
                    collected_msg = []
    for p in procs:
        p.join()
    siz = ntmesh.shape[1]
    if iex:
        return np.array([np.array(iq).flatten(),
                         np.array(id).flatten(),
                         np.array(iex).flatten()])[:, :siz]

    return np.array([np.array(iq).flatten(),
                        np.array(id).flatten()])[:, :siz]


def rectangular_grid(ntmesh): 
    """return speed and torque with a rectangular grid

    arguments:
    ntmesh: speed and torque list generated by the function _generate_mesh()
    """
    n = ntmesh[0, :]
    T = ntmesh[1, :]
    nmesh = [] 
    tmesh = []
    ngrid = np.unique(n)

    npoints = len(np.argwhere(n == ngrid[0]).squeeze())
    if npoints%2: 
        npoints = npoints + 1
    
    tp = []
    tn = []

    for i in ngrid: 
        idx = np.argwhere(n == i).squeeze()
        if len(idx) > 0: 
            tinp = T[idx]
            # search pos. and neg. torque
            idx_pos = np.argwhere(tinp > 0).squeeze()
            idx_neg = np.argwhere(tinp < 0).squeeze()

            # pos. half; motor range
            tmax_pos, tmin_pos = np.amax(tinp[idx_pos]), np.amin(tinp[idx_pos])
            tp = np.linspace(tmin_pos, tmax_pos, int(npoints/2), endpoint=True)

            # neg. half; generator range
            if len(idx_neg) > 0: 
                tmax_neg, tmin_neg = np.amax(tinp[idx_neg]), np.amin(tinp[idx_neg])
                tn = np.linspace(tmin_neg, tmax_neg, int(npoints/2), endpoint=True)
                tmesh += tn.tolist() + tp.tolist()
            else: 
                tmesh += tp.tolist()

            nmesh += npoints*[i]
        else: 
            pass
    return np.array([[i, j] for i, j in zip(nmesh, tmesh)]).T

def _generate_mesh(n, T, nb, Tb, npoints):
    """return speed and torque list for driving/braking speed range

    arguments:
    n: list of speed values in driving mode (1/s)
    T: list of torque values in driving mode (Nm)
    nb: list of speed values in braking mode (1/s)
    Tb: list of torque values in braking mode (Nm)
    npoints: number of values for speed and torque list
    """
    if nb:
        nmax = 0.99*min(max(n), max(nb))
        tmin, tmax = min(Tb), max(T)
        tnum = npoints[1]//2
    else:
        nmax = max(n)
        tmin, tmax = 0, max(T)
        tnum = npoints[1]
    tip = ip.interp1d(n, T)
    if nb and Tb:
        tbip = ip.interp1d(nb, Tb)
    else:
        def tbip(x): return 0

    nxtx = []
    for nx in np.linspace(n[1], nmax, npoints[0]):
        t0 = tbip(nx)
        t1 = tip(nx)
        npnts = max(round((t1-t0) / (tmax-tmin) * tnum), 2)
        if nb:
            a = np.concatenate(
                        (np.linspace(t0, 0.015*tmin, npnts),
                         np.linspace(0.015*tmax, t1, npnts)))
        else:
            a = np.linspace(0.015*tmax, t1, npnts)
        for t in a:
            nxtx.append((nx, t))
    return np.array(nxtx).T


def efficiency_losses_map(eecpars, u1, T, temp, n, npoints=(60, 40),
                          with_mtpv=True, with_mtpa=True, with_pmconst=True,
                          with_tmech=True, driving_only=False,
                          num_proc=0, progress=None, **kwargs):
    """return speed, torque efficiency and losses

    arguments:
    eecpars: (dict) EEC Parameter with
      dicts at different temperatures (or machine object)
    u1: (float) phase voltage (V rms)
    T: (float) starting torque (Nm)
    temp: temperature (°C) (ignored if eecpars is machine objectb)
    n: (float) maximum speed (1/s)
    npoints: (list) number of values of speed and torque
    driving_only: (bool) do not calculate braking speed/torque samples if True
    with_mtpv -- (optional) use mtpv if True (default)
    with_pmconst -- (optional) use pmax if True (default)
    with_mtpa -- (optional) use mtpa if True (default), disables mtpv if False
    with_tmech -- (optional) use friction and windage losses (default)
    num_proc -- (optional) number of parallel processes (default 0)
    progress  -- (optional) custom function for progress logging
    """
    if isinstance(eecpars, dict):
        if isinstance(temp, (list, tuple)):
            xtemp = [temp[0], temp[1]]
        else:
            xtemp = [temp, temp]
        m = create_from_eecpars(xtemp, eecpars)
    else:  # must be an instance of Machine
        m = eecpars
    if isinstance(T, list):
        r = {'T': T, 'n': n}
        rb = {'T': [], 'n': []}
    else:  # calculate speed,torque characteristics
        nmax = n
        nsamples = npoints[0]
        rb = {}
        r = m.characteristics(T, nmax, u1, nsamples=nsamples,
                              with_mtpv=with_mtpv, with_mtpa=with_mtpa,
                              with_pmconst=with_pmconst, with_tmech=with_tmech)  # driving mode
        if driving_only:
            rb['n'] = None
            rb['T'] = None
        elif isinstance(m, (PmRelMachineLdq, SynchronousMachineLdq)):
            if min(m.betarange) >= -np.pi/2:  # driving mode only
                rb['n'] = None
                rb['T'] = None
        elif isinstance(m, (PmRelMachinePsidq, SynchronousMachinePsidq)):
            if min(m.iqrange) >= 0:  # driving mode only
                rb['n'] = None
                rb['T'] = None
        if 'n' not in rb:
            rb = m.characteristics(-T, max(r['n']), u1, nsamples=nsamples,
                                   with_mtpv=with_mtpv, with_mtpa=with_mtpa,
                                   with_pmconst=with_pmconst, with_tmech=with_tmech)  # braking mode
    ntmesh = _generate_mesh(r['n'], r['T'],
                            rb['n'], rb['T'], npoints)

    if kwargs.get('rect_grid', 0):
        ntmesh = rectangular_grid(ntmesh)

    logger.info("total speed,torque samples %d", ntmesh.shape[1])
    if isinstance(m, (PmRelMachine, SynchronousMachine)):
        if num_proc > 1:
            iqd = iqd_tmech_umax_multi(num_proc, ntmesh, m, u1, with_mtpa)
        else:
            class ProgressLogger:
                def __init__(self, nsamples):
                    self.n = 0
                    self.nsamples = nsamples
                    self.num_iv = round(nsamples/15)
                def __call__(self, iqd):
                    self.n += 1
                    if self.n % self.num_iv == 0:
                        logger.info("Losses/Eff Map: %d%%",
                                    round(100*self.n/self.nsamples))
            if progress is None:
                progress = ProgressLogger(ntmesh.shape[1])
            else:
                try:
                    progress.nsamples=ntmesh.shape[1]
                    progress(0)  # To check conformity
                    progress.n = 0
                except:
                    logger.warning("Invalid ProgressLogger given to efficiency_losses_map, using default one!")
                    progress = ProgressLogger(ntmesh.shape[1])
            if with_tmech:
                iqd = np.array([
                    m.iqd_tmech_umax(
                        nt[1],
                        2*np.pi*nt[0]*m.p,
                        u1, log=progress, with_mtpa=with_mtpa)[:-1]
                    for nt in ntmesh.T]).T
            else:
                iqd = np.array([
                    m.iqd_torque_umax(
                        nt[1],
                        2*np.pi*nt[0]*m.p,
                        u1, log=progress, with_mtpa=with_mtpa)[:-1]
                    for nt in ntmesh.T]).T

        beta, i1 = betai1(iqd[0], iqd[1])
        uqd = [m.uqd(2*np.pi*n*m.p, *i)
               for n, i in zip(ntmesh[0], iqd.T)]
        u1 = np.linalg.norm(uqd, axis=1)/np.sqrt(2.0)
        f1 = ntmesh[0]*m.p
    else:
        f1 = []
        u1max = u1
        r = dict(u1=[], i1=[], plfe1=[], plcu1=[], plcu2=[])
        for nx, tq in ntmesh.T:
            wm = 2*np.pi*nx
            w1 = m.w1(u1max, m.psiref, tq, wm)
            f1.append(w1/2/np.pi)
            u1 = m.u1(w1, m.psi, wm)
            r['u1'].append(np.abs(u1))
            i1 = m.i1(w1, m.psi, wm)
            r['i1'].append(np.abs(i1))
            r['plfe1'].append(m.m*np.abs(u1)**2/m.rfe(w1, m.psi))
            i2 = m.i2(w1, m.psi, wm)
            r['plcu1'].append(m.m*np.abs(i1)**2*m.rstat(w1))
            r['plcu2'].append(m.m*np.abs(i2)**2*m.rrot(w1-m.p*wm))

    if isinstance(m, (PmRelMachine, SynchronousMachine)):
        plfe1 = m.iqd_plfe1(*iqd, f1)
        plfe2 = m.iqd_plfe2(iqd[0], iqd[1], f1)
        plmag = m.iqd_plmag(iqd[0], iqd[1], f1)
        plcu1 = m.iqd_plcu1(iqd[0], iqd[1], 2*np.pi*f1)
        plcu2 = m.iqd_plcu2(*iqd)
        tfric = m.tfric
    else:
        plfe1 = np.array(r['plfe1'])
        plfe2 = np.zeros(ntmesh.shape[1])
        plmag = np.zeros(ntmesh.shape[1])
        plcu1 = np.array(r['plcu1'])
        plcu2 = np.array(r['plcu2'])
        iqd = np.zeros(ntmesh.shape)
        u1 = np.array(r['u1'])
        i1 = np.array(r['i1'])
        try:
            if isinstance(eecpars, dict):
                tfric = eecpars['kfric_b']*eecpars['rotor_mass']*30e-3/np.pi
            else:
                tfric = m.tfric
        except KeyError:
            tfric = 0

    plfric = 2*np.pi*ntmesh[0]*tfric
    if not with_tmech:
        ntmesh[1] -= tfric
    pmech = np.array(
        [2*np.pi*nt[0]*nt[1]
         for nt in ntmesh.T])
    ploss = plfe1+plfe2+plmag+plcu1+plcu2+plfric

    eta = []
    for pm, pl in zip(pmech, ploss):
        p1 = pm+pl
        if (p1 <= 0 and pm >= 0) or (p1 >= 0 and pm <= 0):
            e = 0
        elif abs(p1) > abs(pm):
            e = pm / p1
        else:
            e = p1 / pm
        eta.append(e)

    return dict(
        iq=iqd[0].tolist(),
        id=iqd[1].tolist(),
        i1=i1.tolist(),
        u1=u1.tolist(),
        n=ntmesh[0].tolist(),
        T=ntmesh[1].tolist(),
        pmech=pmech.tolist(),
        eta=eta,
        plfe1=plfe1.tolist(),
        plfe2=plfe2.tolist(),
        plmag=plmag.tolist(),
        plcu1=plcu1.tolist(),
        plcu2=plcu2.tolist(),
        plfric=plfric.tolist(),
        losses=ploss.tolist())



'''
    femagtools.amela
    ~~~~~~~~~~~~~~~~

    Calculate Magnet Losses with AMELA
'''
import sys
import json
import pathlib
import logging
from pathlib import Path
import femagtools.nc
import numpy as np
from numpy import pi, sin, cos
import subprocess

#  set logging
logger = logging.getLogger(__name__)

def geometry_id(bndx, bndy):
    '''identify the magnet geometry''' 
    # caculate magnet area 
    xy = 0
    yx = 0
    for i in range(len(bndy)):
        if i < (len(bndy)-1):
            xy += bndx[i]*bndy[i+1]
            yx += bndy[i]*bndx[i+1]
        else:
            xy += bndx[i]*bndy[0]
            yx += bndy[i]*bndx[0]
    area = np.abs(yx-xy)*0.5

    x0, y0 = np.mean(bndx), np.mean(bndy)
    distances = [] 
    for i in range(len(bndx)): 
        for j in range(len(bndy)): 
            dist = np.sqrt((bndx[i] - bndx[j])**2 + (bndy[i] - bndy[j])**2)
            distances.append([dist, bndx[i], bndx[j], bndy[i], bndy[j]])
    
    distances.sort(reverse=True)
    xe = [distances[0][2], distances[2][1], distances[0][1], distances[2][2]]
    ye = [distances[0][4], distances[2][3], distances[0][3], distances[2][4]]

    # dimension 
    dim = np.zeros((3, 5))
    x1, y1 = xe[0], ye[0]

    for i in range(3): 
        x2, y2 = xe[i+1], ye[i+1]
        dim[i, 0] = np.sqrt((x2-x1)**2 + (y2-y1)**2)
        dim[i, 1:3] = [x1, x2]
        dim[i, 3:5] = [y1, y2]

    dim = dim[np.lexsort((dim[:, 2], dim[:, 1], dim[:, 0]))]
    dx12, dy12 = dim[1, 1] - dim[1, 2],  dim[1, 3] - dim[1, 4]
    dx14, dy14 = dim[0, 1] - dim[0, 2],  dim[0, 3] - dim[0, 4]
    
    alp1 = np.arctan2(dy12, dx12)
    alp2 = np.arctan2(dy14, dx14)

    if alp1 < 0: alp1 += 2*np.pi
    if alp2 < 0: alp2 += 2*np.pi
    if alp2 < alp1: 
        alp2 += np.pi
        alpha = (alp1 + (alp2 - np.pi/2))/2
    else: 
        alpha = (alp1 + (alp2 - np.pi/2))/2 + np.pi
    
    wm = dim[1, 0]
    hm = area/wm 

    return dict(wm=wm, 
                hm=hm, 
                x0=x0, 
                y0=y0,
                area=area, 
                alpha=alpha)

def tf(b1, b2, alpha):
    '''Tranformation Matrix'''
    T = np.array([[cos(alpha), sin(alpha)], 
                [-sin(alpha), cos(alpha)]]) 
    if b1.ndim > 1: 
        r = T.dot(((b1.ravel()), (b2.ravel())))
        return [r[0, :].reshape(*b1.shape), 
                r[1, :].reshape(*b1.shape)]
    else: 
        return T.dot(((b1), (b2)))

def transform_coord(geometry, xcp, ycp): 
    '''transform from global coord to local coord'''
    # transformation 
    elcp = tf(b1=np.array(xcp)-geometry['x0'], 
              b2=np.array(ycp)-geometry['y0'], 
              alpha=geometry['alpha'])
    return dict(excpl=elcp[0, :]+geometry['wm']/2, 
                eycpl=elcp[1, :]+geometry['hm']/2, 
                excp=np.array(xcp), 
                eycp=np.array(ycp))

def transform_flux_denstiy(geometry, bx, by): 
    '''transform the magnet flux density to local coordinate system'''
    # transformation 
    bxy = tf(b1=bx, 
             b2=by, 
             alpha=geometry['alpha'])

    # remove DC component
    bxf = np.mean(bxy[0].T - np.mean(bxy[0],axis=1).T,axis=1)
    byf = np.mean(bxy[1].T - np.mean(bxy[1],axis=1).T,axis=1)               
     
    return dict(bxl=bxy[0],
                byl=bxy[1],
                bxf=bxf, 
                byf=byf 
                )

class Amela():
    '''Run Amela Calculation
    Parameters
    ----------
    workdir : str
        working directory of femag calculation
        (The directory that the nc file)
    amela_dir: str (optional)
        amela directory 
    magnet_data : dict
        calculation control
        -- name must be provided
        dict(name='test',
        -- the following parameters are optional
             mur=1.05,
             sigma=625000,
             hm=3,
             wm=20,
             lm=30,
             speed=1000, 
             nsegwid=0, 
             nseglen=0)
    '''

    def __init__(self, workdir, magnet_data, amela_dir=None):

        self.magn = magnet_data
        self.workdir = pathlib.Path(workdir)
        if amela_dir is not None: 
            self.amela_dir = pathlib.Path(amela_dir)
        else: 
            self.amela_dir = self.workdir

        self.jsonfile = []
        if sys.platform == 'win32':
            self.cmd = [str(self.amela_dir / 'AMELA.BAT')]
        else:
            self.cmd = [str(self.amela_dir / 'AMELA')]
        # default batch 
        self.cmd.append('--b')
        # append calc options
        if 'speed' in self.magn:
            self.cmd.append(f"--speed {self.magn['speed']}")
        if 'nsegwid' in self.magn:
            self.cmd.append(f"--nsegwid {self.magn['nsegwid']}")
        if 'nseglen' in self.magn:
            self.cmd.append(f"--nseglen {self.magn['nseglen']}")

    def get_magnet_data(self, ibeta=None):
        '''Extract magnet data from nc file
        Parameters
        ----------
            None
        Returns
        ----------
        pm_data: list of magnet data
        '''
        nc_name = self.workdir / self.magn['name']
        r = femagtools.nc.read(nc_name)

        mag_spels = r.magnet_super_elements()
        spel_key = [i.key for i in mag_spels]
        spel_area = [float(i.area())*1e6 for i in mag_spels]

        pm_elem_key = []
        elem = []
        pm_node_key = [[] for i in range(len(spel_key))]
        bndkey = [[] for i in range(len(spel_key))]
        bx = [[] for i in range(len(spel_key))]
        by = [[] for i in range(len(spel_key))]
        xcp = [[] for i in range(len(spel_key))]
        ycp = [[] for i in range(len(spel_key))]
        bndx = [[] for i in range(len(spel_key))]
        bndy = [[] for i in range(len(spel_key))]
        # prepare data for ialh method
        wm = []
        hm = []
        alpha = []
        x0 = [] 
        y0 = []
        geometry = []
        elcp = []
        bl = []
        # conductivity and permeability of the magnets
        cond = 0
        mur = 0
        # read boundary nodes
        for k, i in enumerate(mag_spels):

            cond = i.conduc
            mur = 1/i.elements[0].reluc[0]
            logger.debug('Magnet: mur=%s, conductivity=%s', mur, cond)

            pm_elem_key.append([j.key for j in i.elements])
            elem.append(len(i.elements))
            for j in i.elements:
                for kk in j.vertices:
                    pm_node_key[k].append(kk.key - 1)

            for bnd in i.nodechains:
                for kk in bnd.nodes:
                    if len(bndkey[k]) > 0:
                        if kk.key != bndkey[k][-1]:
                            bndkey[k].append(kk.key)
                            bndx[k].append(kk.x*1e3)
                            bndy[k].append(kk.y*1e3)
                    else:
                        bndkey[k].append(kk.key)
                        bndx[k].append(kk.x*1e3)
                        bndy[k].append(kk.y*1e3)

                bndkey[k].pop(-1)
                bndx[k].pop(-1)
                bndy[k].pop(-1)
            geo = geometry_id(bndx=bndx[k], bndy=bndy[k])
            geometry.append(geo)
            # necessary?
            hm.append(geo['hm'])
            wm.append(geo['wm'])
            x0.append(geo['x0'])
            y0.append(geo['y0'])
            alpha.append(geo['alpha'])

        # default load angle (input beta I vs Up)
        if ibeta is not None:
            indx = ibeta
        else:
            num_cases = r.el_fe_induction_1.shape[3] - 1
            if 'loadcase' in self.magn:
                indx = self.magn['loadcase'] - 1
            else:
                indx = num_cases
            if indx == 3:
                indx = num_cases  # avoid error

        # stationary case, no rotation
        poles = 0
        try:
            poles = r.num_poles
        except:
            pass

        # read mesh and flux density
        for i in range(len(spel_key)):
            bx[i].append(np.zeros((elem[i], len(r.pos_el_fe_induction))))
            by[i].append(np.zeros((elem[i], len(r.pos_el_fe_induction))))
            for index, j in enumerate(mag_spels[i].elements):
                xcp[i].append(float(j.center[0]*1e3))
                ycp[i].append(float(j.center[1]*1e3))
                theta = np.arctan2(float(j.center[1]), float(j.center[0]))
                fd = r.flux_density(j, 0, indx)
                if poles == 0:
                    bx[i][0][index, :] = fd['bx']
                    by[i][0][index, :] = fd['by']
                else:
                    bx[i][0][index, :] = fd['bx']*np.cos(theta) - \
                        fd['by']*np.sin(theta)
                    by[i][0][index, :] = fd['bx']*np.sin(theta) + \
                        fd['by']*np.cos(theta)

            elcp.append(transform_coord(geometry[i], xcp[i], ycp[i]))
            bl.append(transform_flux_denstiy(geometry[i], bx[i][0], by[i][0]))


        if poles == 0:
            freq = self.magn.get('f', r.speed)
            time_vec = np.linspace(0, 1/freq, len(r.pos_el_fe_induction))
            pos = dict(time=time_vec.tolist(),
                       freq=freq,
                       t=float(1/freq))
            # reset num.poles
            poles = 1
        else:
            rpm = self.magn.get('speed', r.speed)
            ag_sim = r.pos_el_fe_induction[-1] - r.pos_el_fe_induction[0]
            pos = dict(phi=(r.pos_el_fe_induction*180/pi).tolist(),
                       speed=rpm,
                       t=float(60/rpm*ag_sim/360))  # TODO
        # prep dictionary for the loss calculation
        pm_data = []
        for i in range(len(spel_key)):
            pm_data.append(dict(name='pm_data_se' + str(spel_key[i]),
                                hm=self.magn.get('hm', hm[i]),
                                wm=self.magn.get('wm', wm[i]),
                                lm=self.magn.get('lm', r.arm_length*1e3),
                                alpha=alpha[i],
                                ls=r.arm_length*1e3,
                                sigma=float(self.magn.get('sigma', cond)),
                                mur=float(self.magn.get('mur', mur)),
                                loadcase=self.magn.get('loadcase', indx),
                                numpoles=poles,
                                nodes=dict(),
                                elements=dict(),
                                bndkeys=bndkey[i],
                                bndx=[float(c) for c in bndx[i]],
                                bndy=[float(c) for c in bndy[i]],
                                bl=bl[i],
                                elcp=elcp[i],
                                area=spel_area[i], 
                                spel_key=spel_key[i]))
            pm_data[i].update(pos)

        for k in range(len(pm_node_key)):
            for i, j in enumerate(pm_node_key[k]):
                pm_data[k]['nodes'][str(j + 1)] = dict(
                    key=int(j + 1),
                    x=float(r.nodes[j].x*1e3),
                    y=float(r.nodes[j].y*1e3),
                    Az=float(r.nodes[j].vpot[0])
                )
            for i, j in enumerate(pm_elem_key[k]):
                pm_data[k]['elements'][str(j)] = dict(
                    key=int(j + 1),
                    xcp=xcp[k][i],
                    ycp=ycp[k][i],
                    Bx=bx[k][0][i, :].tolist(),
                    By=by[k][0][i, :].tolist()
                )

        if len(mag_spels) / r.poles_sim > 1:
            return pm_data
        else:
            return [pm_data[0]]

    def get_magnet_data_all(self, num_op): 
        '''get all magnet data for all loadcases'''
        pm_data = []
        for i in num_op:
            pmd = self.get_magnet_data(ibeta=i)
            pm_data.append(pmd)
        return pm_data

    def export_json(self, pm_data):
        '''Export magnet data to json files
        Parameters
        ----------
        pm_data: list of magnet data
        Returns
        ----------
        None
        ----------
        '''
        pm_dir = self.amela_dir / self.magn['name']
        pm_dir.mkdir(exist_ok=True)
        for i in pm_data:
            filename = (pm_dir / i['name']).with_suffix('.json')
            self.jsonfile.append(str(filename))  # for the future use
            # pop out non necessary data
            i.pop('bl')
            i.pop('elcp')
            with filename.open('w') as f:
                json.dump(i, f)
            logger.info('Exporting %s ...', i['name'])

    def read_loss(self, pm_data):
        '''Read magnet losses
        Parameters
        ----------
            pm_data : dict
        Returns
        ----------
            losses : dict
        ----------
        '''
        losses = {}
        for i in range(len(pm_data)):
            dirname = self.magn['name'] + '/' + pm_data[i]['name'] + \
                '/EClosses.tab'
            result_name = self.amela_dir / dirname
            with result_name.open() as f:
                data = np.loadtxt(f)
                total_loss = np.mean(data[0:-1, -1]) 
                losses[pm_data[i]['name']] = {"loss_data":data, "total_loss": total_loss}
                logger.info("Magnet losses in superelement %s is %.3f W",
                            pm_data[i]['name'].split('se')[-1], total_loss)
        return losses

    def __call__(self, ialh=False):
        '''Run amela calculation
        Parameters
        ----------
            None
        Returns
        ----------
            losses : dict
        ----------
        '''
        # get magnet data
        r = self.get_magnet_data()
        # export to json
        self.export_json(r)
        # run amela 
        calc_method = 'IALH' if ialh else '3DI'
        cmd = self.cmd + ['--calc', calc_method, self.magn['name'] + '/']
        log_file = self.amela_dir / 'amela.out'
        logger.info("Calculating magnet losses with AMELA ...")
        with log_file.open('w') as output:
            subprocess.run(cmd, stdout=output)
        return self.read_loss(r)



"""
    femagtools.vbf
    ~~~~~~~~~~~~~~

    Manage VBF magnetizing curve data files



"""
import sys
import femagtools.losscoeffs as lc
import logging

logger = logging.getLogger(__name__)

fo = 50.
Bo = 1.5


class Reader(object):

    def __init__(self, filename, filecontent=None):
        self.losses = {}
        # filecontent is used
        if filecontent:
            content = [l.strip() for l in filecontent.split('\n')]
        else:
            # only filename
            with open(filename) as f:
                content = [l.strip() for l in f.readlines()]

        if content:
            self.losses['name'] = content[0]
            self.losses['fo'], self.losses['Bo'] = [float(s)
                                                    for s in content[1].split()]
            # Ignore the next line
            self.losses['f'] = [float(s) for s in content[3].split()]
            self.losses['B'] = []
            self.losses['pfe'] = []
            for l in content[4:]:
                values = [float(s) for s in l.strip().split()]
                if len(values) > 1:
                    self.losses['B'].append(values[0])
                    self.losses['pfe'].append(
                        [v if v > 0 else None for v in values[1:]])
        logger.info("%s fmax %5.1f Bmax %3.2f", filename,
                    max(self.losses['f']), max(self.losses['B']))

    def __getitem__(self, index):
        return self.losses[index]
    
    def getLossValues(self):
        return self.losses


def read(filename, filecontent=None):
    """read VBF file and return dict of content"""
    vbf = Reader(filename, filecontent=filecontent)
    return vbf.getLossValues()
    

if __name__ == "__main__":
    import matplotlib.pylab as pl
    import femagtools.plot
    
    if len(sys.argv) == 2:
        filename = sys.argv[1]
    else:
        filename = sys.stdin.readline().strip()

    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s')
    losses = read(filename)
    print(losses)
    
    femagtools.plot.felosses(losses,
                             lc.fitjordan(
                                 losses['f'],
                                 losses['B'],
                                 losses['pfe'],
                                 losses['Bo'],
                                 losses['fo']),
                             title=filename, log=False)
    pl.show()



# -*- coding: utf-8 -*-
"""
    femagtools.gmsh
    ~~~~~~~~~~~~~~~~

    Handle gmsh models


"""
import numpy as np
import logging

class Gmsh(object):
    def __init__(self, mshfile):
        import meshio
        self.m = meshio.read(mshfile)
        self.r = np.linalg.norm(self.m.points, axis=1)
        self.phi = [np.arctan2(p[1], p[0]) for p in self.m.points]

    def get_section_angles(self):
        return np.min(self.phi), np.max(self.phi)

    def get_section_radius(self):
        return np.min(self.r), np.max(self.r)

    def get_subregions(self):
        return self.m.field_data.keys()

    def get_points(self, srname):
        """return x,y coordinates of all points in subregion srname"""
        srid = self.m.field_data[srname][0]
        trids = [i for i, k in enumerate(
            self.m.cell_data['gmsh:physical'][0]) if k == srid]
        quids = [i for i, k in enumerate(
            self.m.cell_data['gmsh:physical'][1]) if k == srid]
        p = np.unique(np.concatenate((self.m.cells[0].data[trids].flatten(), 
                                      self.m.cells[1].data[quids].flatten())))
        return self.m.points[p]
    
    def get_location(self, srname):
        """return x,y coordinates of first element (triangle, quad) in subregion srname"""
        srid = self.m.field_data[srname][0]
        eids = [i for i, k in enumerate(self.m.cell_data['gmsh:physical'][0]) if k == srid]
        if eids:
            return (np.sum(self.m.points[self.m.cells[0].data[eids[0]]], axis=0)/3)[:2]        
        eids = [i for i, k in enumerate(self.m.cell_data['gmsh:physical'][1]) if k == srid]
        if eids:
            return (np.sum(self.m.points[self.m.cells[1].data[eids[0]]], axis=0)/4)[:2]        
        raise ValueError("subregion '{}' not found".format(srname))

    def get_corners(self, srname):
        """return corner points in counterclockwise order"""
        p = self.get_points(srname)
        corner = []
        # 1: lower left
        for p1 in p:
            x = []
            for p2 in p:
                if p1[0] > p2[0]: 
                    x.append(p2)
                    break
            if not x:
                corner.append(p1[:2])
        # 2: lower right
        for p1 in p:
            x = []
            for p2 in p:
                if p1[1] > p2[1]: 
                    x.append(p2)
                    break
            if not x:
                corner.append(p1[:2])
        # 3: upper right
        for p1 in p:
            x = []
            for p2 in p:
                if p1[0] < p2[0]: 
                    x.append(p2)
                    break
            if not x:
                corner.append(p1[:2])
        # 4: upper left
        for p1 in p:
            x = []
            for p2 in p:
                if p1[1] < p2[1]: 
                    x.append(p2)
                    break
            if not x:
                corner.append(p1[:2])
        return corner

    def get_axis_angle(self, srname):
        """returns angle of axis in degrees"""
        corners = self.get_corners(srname)
        logging.debug("Corners %s of '%s'", corners, srname)
        l = corners[3], corners[2]
        alfa = np.arctan2(l[0][1] - l[1][1],
                          l[0][0] - l[1][0])/np.pi*180
                          
        logging.debug("Line l %s angle %s", l, alfa)
        return alfa
    
if __name__ == "__main__":
    import femagtools.isa7
    import logging
    import os
    import sys
    
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s')
    if len(sys.argv) == 2:
        filename = sys.argv[1]
    else:
        filename = sys.stdin.readline().strip()

    isa = femagtools.isa7.read(filename)

    basename = os.path.splitext(os.path.basename(filename))[0]
    with open(basename + '.msh', 'w') as f:
        f.write('\n'.join(isa.msh()))



"""
  femagtools.dxfsl.conv

  entry point for DXF to FSL conversion

"""
import sys
import os
import io
import femagtools
from femagtools.dxfsl.converter import convert
import argparse
import logging
import logging.config

logger = logging.getLogger(__name__)


def main():
    argparser = argparse.ArgumentParser(
        description='Process DXF file and create a plot or FSL file.')
    argparser.add_argument('dxfile',
                           help='name of DXF file')
    argparser.add_argument('--inner',
                           help='name of inner element',
                           dest='inner',
                           default='inner')
    argparser.add_argument('--outer',
                           help='name of outer element',
                           dest='outer',
                           default='outer')
    argparser.add_argument('--rotor',
                           help='rotor without airgap in/out',
                           dest='rotor',
                           default='')
    argparser.add_argument('--stator',
                           help='stator without airgap in/out',
                           dest='stator',
                           default='')
    argparser.add_argument('--sympart',
                           help='forced symmetry part',
                           dest='sym_part',
                           type=int,
                           default=0)
    argparser.add_argument('-a', '--airgap',
                           help='correct airgap',
                           dest='airgap',
                           type=float,
                           default=0.0)
    argparser.add_argument('--airgap2',
                           help='correct airgap',
                           dest='airgap2',
                           type=float,
                           default=0.0)
    argparser.add_argument('-t', '--symtol',
                           help='absolut tolerance to find symmetry axis',
                           dest='sym_tolerance',
                           type=float,
                           default=0.001)
    argparser.add_argument('--mindist',
                           help='minimal distance of spline control-points',
                           dest='mindist',
                           type=float,
                           default=0.01)
    argparser.add_argument('--rtol',
                           help='relative tolerance (pickdist)',
                           dest='rtol',
                           type=float,
                           default=1e-03)
    argparser.add_argument('--atol',
                           help='absolut tolerance (pickdist)',
                           dest='atol',
                           type=float,
                           default=0.005)
    argparser.add_argument('--da',
                           help='distance airgap',
                           dest='da',
                           type=float,
                           default=0.0)
    argparser.add_argument('--dy',
                           help='distance yoke',
                           dest='dy',
                           type=float,
                           default=0.0)
    argparser.add_argument('-s', '--split',
                           help='split intersections',
                           dest='split',
                           action="store_true")
    argparser.add_argument('-p', '--plot',
                           help='show plots',
                           dest='show_plots',
                           action="store_true")
    argparser.add_argument('--areas',
                           help='show all areas',
                           dest='show_areas',
                           action="store_true")
    argparser.add_argument('--id',
                           help='write id of areas',
                           dest='write_id',
                           action="store_true")
    argparser.add_argument('-f', '--fsl',
                           help='create fsl',
                           dest='write_fsl',
                           action="store_true")
    argparser.add_argument('-v', '--view',
                           help='show a view only',
                           dest='view',
                           action="store_true")
    argparser.add_argument('-k', '--korr',
                           help='show a view with korrections',
                           dest='view_korr',
                           action="store_true")
    argparser.add_argument('--png',
                           help='write png-file only',
                           dest='write_png',
                           action="store_true")
    argparser.add_argument('-d', '--debug',
                           help='print debug information in logfile',
                           dest='debug',
                           action="store_true")
    argparser.add_argument('-l', '--log',
                           help='print information in logfile',
                           dest='debug',
                           action="store_true")
    argparser.add_argument('--version',
                           help='show version of some packages',
                           dest='version',
                           action="store_true")
    argparser.add_argument('--debugger',
                           help='print debug information in logfile',
                           dest='debugger',
                           action="store_true")

    args = argparser.parse_args()

    logfilename = None
    loglevel = logging.INFO
    if args.debug:
        loglevel = logging.DEBUG
        logfilename = 'debugger.log'
        print("see log-messages in {}".format(logfilename))

    logging.basicConfig(level=loglevel,
                        format='%(asctime)s %(message)s',
                        filename=logfilename,
                        filemode='w')

    if args.version:
        logger.info("femagtools version: %s", femagtools.__version__)
        try:
            import networkx as nx
            logger.info("networkx version: %s", nx.__version__)
        except ImportError:  # ModuleNotFoundError:
            logger.info("networkx version: <networkx not available>")
        try:
            import matplotlib
            logger.info("matplotlib version: %s", matplotlib.__version__)
        except ImportError:  # ModuleNotFoundError:
            logger.info("matplotlib version: <matplotlib not available>")
        logger.info("Python: %s", sys.version)
        sys.exit(0)

    if args.airgap > 0.0:
        if args.airgap2 > 0.0:
            logger.info("Airgap is set from {} to {}"
                        .format(args.airgap, args.airgap2))
        else:
            logger.info("Airgap is set to {}".format(args.airgap))

    part = ()
    if args.stator:
        if args.rotor:
            logger.error("Stator or Rotor expected")
            sys.exit(1)
        part = ('stator', args.stator)
    elif args.rotor:
        part = ('rotor', args.rotor)
    if part:
        if args.airgap:
            logger.info('airgap in stator or rotor not possible')
            sys.exit(1)
        args.airgap = -1  # no airgap
        if part[1] not in ('in', 'out'):
            logger.info('{} has to be defined in/out'.format(part[0]))
            sys.exit(1)
    if args.sym_part > 0:
        if not part:
            logger.error("Argument sympart only with Stator or Rotor")
            sys.exit(1)
        if args.sym_part not in (3, 4, 6, 8):
            logger.error("Argument sympart not in (3, 4, 6, 8)")
            sys.exit(1)

    if not args.write_fsl:
        if not (args.show_plots or args.show_areas or args.view):
            args.write_fsl = True

    res = convert(args.dxfile,  # DXF-Filename
                  rtol=args.rtol,    # relative pickdist toleranz
                  atol=args.atol,    # absolute pickdist toleranz
                  symtol=args.sym_tolerance,
                  sympart=args.sym_part,
                  mindist=args.mindist,
                  split=args.split,
                  inner_name=args.inner,
                  outer_name=args.outer,
                  part=part,
                  airgap=args.airgap,
                  airgap2=args.airgap2,
                  da=args.da,  # distance airgap
                  dy=args.dy,  # distance yoke
                  view_only=args.view,
                  view_korr=args.view_korr,
                  show_plots=args.show_plots,
                  show_areas=args.show_areas,
                  write_fsl=args.write_fsl,
                  write_png=args.write_png,
                  write_id=args.write_id,
                  debug_mode=args.debugger)

    if args.write_fsl:
        if res is not None:
            basename = os.path.basename(args.dxfile).split('.')[0]
            with io.open(basename + '.fsl', 'w', encoding='utf-8') as f:
                f.write('\n'.join(res['fsl']))


if __name__ == "__main__":
    loglevel = logging.INFO

    main()



# -*- coding: utf-8 -*-
"""
  femagtools.dxfsl.area
  ~~~~~~~~~~~~~~~~~~~~~

  areas are regions surrounded by a collection of shapes

  Authors: Ronald Tanner, beat Holm
"""
from __future__ import print_function
import sys
import numpy as np
import networkx as nx
import logging
from .functions import less_equal, less, greater_equal, greater
from .functions import distance, alpha_angle, alpha_line, min_angle, max_angle
from .functions import point, line_m, line_n, intersect_point, points_are_close
from .functions import middle_angle, part_of_circle, is_same_angle
from .functions import area_size
from .shape import Element, Shape, Line, Arc, Circle, is_Circle

logger = logging.getLogger('femagtools.area')


#############################
#            Area           #
#############################

area_number = 0


class Area(object):
    def __init__(self, area, center, sym_tolerance):
        self.area = area
        self.type = 0  # material
        self.phi = 0.0
        self.min_angle = 0.0
        self.max_angle = 0.0
        self.min_air_angle = 0.0
        self.max_air_angle = 0.0
        self.close_to_ag = False
        self.close_to_ag_startcorner = False
        self.close_to_ag_endcorner = False
        self.close_to_startangle = False
        self.close_to_endangle = False
        self.mag_rectangle = False
        self.min_dist = 99999.0
        self.max_dist = 0.0
        self.height = 0.0
        self.alpha = 0.0
        self.count = 1
        self.equal_areas = []
        self.delta = 0.0
        self.start = 0.0
        self.sym_startangle = 0.0
        self.sym_endangle = 0.0
        self.sym_type = 0
        self.symmetry = 0
        self.sym_tolerance = sym_tolerance
        self.calc_signature(center)
        self.surface = 0.0
        global area_number
        area_number += 1
        self.id = area_number
        self.areas_inside = {}

    def identifier(self):
        return "{}-{}".format(self.id, self.type)

    def get_id(self):
        return self.id

    def number_of_elements(self):
        return len(self.area)

    def elements(self):
        return self.area

    def list_of_nodes(self):
        if len(self.area) < 2:
            return

        e0 = self.area[0]
        e1 = self.area[1]
        try:
            if e1.get_node_number(e0.n1, override=True) == 0:
                nx = e0.n2
            else:
                nx = e0.n1
            yield nx

            for e1 in self.area[1:]:
                if e1.get_node_number(nx) == 1:
                    nx = e1.n2
                else:
                    nx = e1.n1
                yield nx
        except ValueError as e:
            logger.error("list_of_nodes(): FATAL ERROR: %s", e)
            return
        except Exception as e:
            return

    def list_of_elements(self):
        if len(self.area) < 2:
            return

        e0 = self.area[0]
        e1 = self.area[1]
        try:
            if e1.get_node_number(e0.n1, override=True) == 0:
                n1 = e0.n1
                n2 = e0.n2
            else:
                n1 = e0.n2
                n2 = e0.n1
            yield n1, n2, e0

            for e1 in self.area[1:]:
                if e1.get_node_number(n2) == 1:
                    n1 = e1.n1
                    n2 = e1.n2
                else:
                    n1 = e1.n2
                    n2 = e1.n1
                yield n1, n2, e1
        except ValueError as e:
            logger.error("list_of_elements(): FATAL ERROR: %s", e)
            return
        except Exception as e:
            return

    def has_fsl(self):
        try:
            return self.fsl
        except AttributeError:
            return True

    def list_of_equal_edges(self, a):
        for e1 in self.area:
            for e2 in a.area:
                if e1.n1 == e2.n1 and e1.n2 == e2.n2:
                    yield e1

    def virtual_nodes(self, render=False):
        if len(self.area) < 2:
            return

        prev_nodes = [n for n in self.area[0].get_nodes(parts=64,
                                                        render=render)]
        next_nodes = [n for n in self.area[1].get_nodes(parts=64,
                                                        render=render)]
        if points_are_close(prev_nodes[0], next_nodes[0], 1e-03, 1e-01):
            prev_nodes = prev_nodes[::-1]
        elif points_are_close(prev_nodes[0], next_nodes[-1], 1e-03, 1e-01):
            prev_nodes = prev_nodes[::-1]
            next_nodes = next_nodes[::-1]
        elif points_are_close(prev_nodes[-1], next_nodes[-1], 1e-03, 1e-01):
            next_nodes = next_nodes[::-1]
        elif not points_are_close(prev_nodes[-1], next_nodes[0], 1e-03, 1e-01):
            assert(False)
        last_point = next_nodes[-1]
        for n in prev_nodes:
            yield n
        for n in next_nodes:
            yield n

        for e in self.area[2::]:
            next_nodes = [n for n in e.get_nodes(parts=64, render=render)]

            if points_are_close(next_nodes[-1], last_point, 1e-03, 1e-01):
                next_nodes = next_nodes[::-1]
            for n in next_nodes:
                yield n
            last_point = next_nodes[-1]

    def legend(self):
        if self.type == 1:
            return 'Iron'
        if self.type == 2:
            return 'Windings'
        if self.type == 3 or self.type == 4:
            return 'Magnet'
        if self.type == 5:
            return 'Yoke'
        if self.type == 6:
            return 'Tooth'
        if self.type == 10:
            return 'Shaft'
        return ''

    def name(self):
        if self.type == 1:
            return 'Iron'
        if self.type == 2:
            return 'Wndg'
        if self.type == 3 or self.type == 4:
            return 'Mag'
        if self.type == 5:
            return 'StJo'
        if self.type == 6:
            return 'StZa'
        if self.type == 10:
            return 'Shft'
        return ''

    def color(self):
        if self.type == 1:
            return 'cyan'
        if self.type == 2:
            return 'green'
        if self.type == 3 or self.type == 4:
            return 'red'
        if self.type == 5:
            return 'cyan'
        if self.type == 6:
            return 'skyblue'
        if self.type == 10:
            return 'lightgrey'
        return 'white'

    def color_alpha(self):
        if self.type == 1:
            return 0.3
        if self.type == 2:
            return 1.0
        if self.type == 3 or self.type == 4:
            return 1.0
        if self.type == 5:
            return 0.5
        if self.type == 6:
            return 1.0
        if self.type == 10:
            return 0.8
        return 1.0

    def is_iron(self):
        return self.type == 1 or self.type == 5 or self.type == 6

    def is_stator_iron_yoke(self):
        return self.type == 5

    def is_stator_iron_tooth(self):
        return self.type == 6

    def is_rotor_iron(self):
        return self.type == 1

    def is_winding(self):
        return self.type == 2

    def is_magnet(self):
        return self.type == 3 or self.type == 4

    def is_shaft(self):
        return self.type == 10

    def is_air(self):
        return self.type == 0

    def set_type(self, t):
        self.type = t

    def calc_signature(self, center):
        if not self.area:
            return

        s = self.area[0]
        mm_angle = s.minmax_angle_from_center(center)
        self.min_angle = mm_angle[0]
        self.max_angle = mm_angle[1]

        for s in self.area:
            mm_dist = s.minmax_from_center(center)
            self.min_dist = min(self.min_dist, mm_dist[0])
            self.max_dist = max(self.max_dist, mm_dist[1])
            self.height = self.max_dist - self.min_dist

            mm_angle = s.minmax_angle_from_center(center)
            self.min_angle = min_angle(self.min_angle, mm_angle[0])
            self.max_angle = max_angle(self.max_angle, mm_angle[1])

        self.alpha = round(alpha_angle(self.min_angle, self.max_angle), 3)

    def minmax_angle_dist_from_center(self, center, dist):
        circ = Circle(Element(center=center, radius=dist))
        s = self.area[0]
        my_min_angle = self.max_angle
        my_max_angle = self.min_angle
        mm_angle = None
        for s in self.area:
            mm_angle = s.minmax_angle_dist_from_center(my_min_angle,
                                                       my_max_angle,
                                                       center, circ)
            if mm_angle:
                my_min_angle = min_angle(my_min_angle, mm_angle[0])
                my_max_angle = max_angle(my_max_angle, mm_angle[1])
        return (my_min_angle, my_max_angle)

    def is_inside(self, area):
        if less_equal(area.min_dist, self.min_dist):
            return False
        if greater_equal(area.max_dist, self.max_dist):
            return False
        if less_equal(area.min_angle, self.min_angle):
            return False
        if greater_equal(area.max_angle, self.max_angle):
            return False
        return True

    def is_touching(self, area):
        for n in self.list_of_nodes():
            x = [p for p in area.list_of_nodes() if points_are_close(n, p)]
            if x:
                return True
        return False

    def is_touching_both_sides(self):
        return (self.close_to_startangle and self.close_to_endangle)

    def has_connection(self, geom, a, ndec):
        assert(self.area)
        assert(a.area)
        n1 = self.area[0].node1(ndec)
        if not geom.g.has_node(n1):
            n = geom.find_nodes(n1)
            if not n:
                logger.warn("FATAL: node {} not available".format(n1))
                return False
            n1 = n[0]

        n2 = a.area[0].node2(ndec)
        if not geom.g.has_node(n2):
            n = geom.find_nodes(n2)
            if not n:
                logger.warn("FATAL: node {} not available".format(n2))
                return False
            n2 = n[0]

        try:
            return nx.has_path(geom.g, n1, n2)
        except nx.NetworkXError:
            logger.warn("has_path() failed")
            return False

    def get_lowest_gap_list(self, a, center, radius, rightangle, leftangle):
        gap_list = []
        if a.get_id() < self.get_id():
            dist_id = '{}-{}'.format(a.get_id(), self.get_id())
        else:
            dist_id = '{}-{}'.format(self.get_id(), a.get_id())

        for p1 in self.list_of_nodes():
            for p2 in a.list_of_nodes():
                d = distance(p1, p2)
                gap_list.append((d, (p1, p2), dist_id, a.get_id()))

        d, p1, p2 = a.get_nearest_point(center, radius, rightangle)
        gap_list.append((d, (p1, p2), dist_id, a.get_id()))
        d, p1, p2 = a.get_nearest_point(center, radius, leftangle)
        gap_list.append((d, (p1, p2), dist_id, a.get_id()))
        gap_list.sort()
        return [gap_list[0]]

    def get_nearest_point(self, center, radius, angle):
        axis_p = point(center, radius, angle)
        axis_m = line_m(center, axis_p)
        axis_n = line_n(center, axis_m)

        the_area_p = None
        the_axis_p = None
        dist = 99999
        for n in self.list_of_nodes():
            p = intersect_point(n, center, axis_m, axis_n)
            d = distance(n, p)
            if d < dist:
                dist = d
                the_area_p = n
                the_axis_p = p

        return (dist,
                (the_axis_p[0], the_axis_p[1]),
                (the_area_p[0], the_area_p[1]))

    def is_equal(self, a, sym_tolerance):
        if sym_tolerance > 0.0:
            if np.isclose(round(self.min_dist, 4),
                          round(a.min_dist, 4),
                          1e-03, sym_tolerance) and \
               np.isclose(round(self.max_dist, 4),
                          round(a.max_dist, 4),
                          1e-03, sym_tolerance) and \
               np.isclose(round(self.alpha, 3),
                          round(a.alpha, 3),
                          1e-02, 0.01):
                return True
        else:
            if np.isclose(round(self.min_dist, 2),
                          round(a.min_dist, 2)) and \
               np.isclose(round(self.max_dist, 2),
                          round(a.max_dist, 2)) and \
               np.isclose(round(self.alpha, 3),
                          round(a.alpha, 3), 1e-02, 0.001):
                return True
        return False

    def is_identical(self, area):
        if np.isclose(self.min_dist, area.min_dist) and \
           np.isclose(self.max_dist, area.max_dist) and \
           np.isclose(self.alpha, area.alpha) and \
           np.isclose(self.min_angle, area.min_angle) and \
           np.isclose(self.max_angle, area.max_angle):
            return True
        return False

    def increment(self, a):
        if self.is_identical(a):
            return

        for area in self.equal_areas:
            if area.is_identical(a):
                return

        self.count += 1
        self.equal_areas.append(a)

    def set_delta(self):
        logger.debug("begin set_delta of {}".format(self.id))
        self.delta = 0.0
        self.symmetry = 0

        if len(self.equal_areas) < 2:
            # Mit zwei Objekten lässt sich das Teil nur noch halbieren. Das
            # wird zum Schluss sowieso versucht.
            logger.debug("end set_delta: zuwenig Gleiche")
            return

        sorted_areas = []
        sorted_areas.append((self.min_angle, self))
        for a in self.equal_areas:
            sorted_areas.append((a.min_angle, a))
        sorted_areas.sort()

        delta = {}
        prev_angle = sorted_areas[0][0]
        for angle, area in sorted_areas[1:]:
            d = round(alpha_angle(prev_angle, angle), 2)
            if d in delta:
                delta[d] += 1
            else:
                delta[d] = 1
            prev_angle = angle

        delta_sorted = list([v, k] for (k, v) in delta.items())
        logger.debug(" - delta: {}".format(delta_sorted))

        if len(delta_sorted) == 1:
            # simple case: all have the same angle
            self.delta = alpha_angle(sorted_areas[0][1].min_angle,
                                     sorted_areas[1][1].min_angle)
            self.start = middle_angle(sorted_areas[0][1].max_angle,
                                      sorted_areas[1][1].min_angle)
            self.sym_type = 3
            self.symmetry = part_of_circle(0.0, self.delta, 1)
            logger.debug("end set_delta: simple case")
            return

        logger.debug("end set_delta: {} deltas, {} areas"
                     .format(len(delta_sorted), len(self.equal_areas)))

        if len(delta_sorted) > 2:
            # Mehr als 2 Winkel untersuchen wir (noch) nicht. Wir brechen
            # die Suche nach dem richtigen Winkel ab.
            logger.debug("end set_delta: zuviele Winkel")
            return

        # Bei 2 verschiedenen Winkeln werden die näher beieinander liegenden
        # Objekte zusammen genommen.

        if len(self.equal_areas) < 4:
            # Wenn nicht mehr als 4 Objekte vorhanden sind, brechen wir auch
            # ab.
            logger.debug("end set_delta: zuwenig areas")
            return

        if np.isclose(delta_sorted[1][1],
                      delta_sorted[0][1]*2, atol=0.01):
            # Lets hope we have unreqognised areas inbetween
            percent = 1.0
        else:
            percent = delta_sorted[0][0] / (len(self.equal_areas)+1)

        if percent > 0.75:
            # lets assume we only have one angle
            self.delta = alpha_angle(sorted_areas[0][1].min_angle,
                                     sorted_areas[1][1].min_angle)
            self.start = middle_angle(sorted_areas[0][1].max_angle,
                                      sorted_areas[1][1].min_angle)
            self.sym_type = 2
            self.symmetry = part_of_circle(0.0, self.delta, 1)
            logger.debug("end set_delta: {} Prozent gleiche deltas"
                         .format(percent))
            return

        # Lets hope the distances are changing
        self.delta = alpha_angle(sorted_areas[0][1].min_angle,
                                 sorted_areas[2][1].min_angle)
        self.sym_type = 1
        self.symmetry = part_of_circle(0.0, self.delta, 1)
        delta_1 = alpha_angle(sorted_areas[0][1].min_angle,
                              sorted_areas[1][1].min_angle)
        delta_2 = alpha_angle(sorted_areas[1][1].min_angle,
                              sorted_areas[2][1].min_angle)

        if np.isclose(delta_1, delta_2):
            # Hm. the distances are not changing
            self.delta = 0.0
            logger.debug("end set_delta: the distances are not changing")
            return

        if delta_1 < delta_2:
            self.start = middle_angle(sorted_areas[1][1].max_angle,
                                      sorted_areas[2][1].min_angle)
        else:
            self.start = middle_angle(sorted_areas[0][1].max_angle,
                                      sorted_areas[1][1].min_angle)
        logger.debug("end set_delta: delta wechselt: delta={}"
                     .format(self.delta))

    def symmetry_lines(self, startangle, endangle):
        logger.debug("begin symmetry_lines of {} ({}, {})"
                     .format(self.id,
                             startangle,
                             endangle))
        if less_equal(endangle, startangle):
            endangle += 2*np.pi

        angle = self.start
        while less(angle, startangle):
            angle += self.delta
        while greater(angle, startangle+self.delta):
            angle -= self.delta

        # Damit man anschliessend ohne Umstände schneiden kann.
        self.sym_startangle = angle
        self.sym_endangle = angle + self.delta
        logger.debug(" - delta: {}, sym start: {}, end: {}"
                     .format(self.sym_startangle,
                             self.delta,
                             self.sym_endangle))
        while angle < endangle:
            yield angle
            angle += self.delta
        logger.debug("end symmetry_lines")

    def minmax(self):
        mm = [99999, -99999, 99999, -99999]

        for e in self.area:
            n = e.minmax()
            mm[0] = min(mm[0], n[0])
            mm[1] = max(mm[1], n[1])
            mm[2] = min(mm[2], n[2])
            mm[3] = max(mm[3], n[3])
        return mm

    def intersect_line(self, line):
        for e in self.area:
            if e.intersect_line(line):
                return True
        return False

    def is_point_inside(self, pt):
        for e in self.area:
            if e.is_point_inside(pt, include_end=True):
                return True
        return False

    def get_best_point_inside(self, geom):
        mm = self.minmax()
        px1 = mm[0]-5
        px2 = mm[1]+5

        y_dist = mm[3] - mm[2]
        step = y_dist / 6
        y_list = np.arange(mm[2] + step*0.3, mm[3] - step*0.3, step)

        lines = []
        for y in y_list:
            p1 = (px1, y)
            p2 = (px2, y)
            line = Line(Element(start=p1, end=p2))
            lines.append({'line': line,
                          'pts': [],
                          'y': y,
                          'x': []})

        for e in self.area:
            points = []
            for line in lines:
                line['pts'] += e.intersect_line(line['line'],
                                                geom.rtol,
                                                geom.atol,
                                                True)
        for line in lines:
            x_sorted = [p[0] for p in line['pts']]
            x_sorted.sort()
            if x_sorted:
                line['start_x'] = x_sorted[0]
                line['end_x'] = x_sorted[-1]

        for e in geom.elements(Shape):
            for line in lines:
                if line.get('start_x', None) is None:
                    continue
                points = e.intersect_line(line['line'],
                                          geom.rtol,
                                          geom.atol,
                                          True)

                for p in points:
                    if greater(p[0],
                               line['start_x'],
                               rtol=1e-8):
                        if less(p[0],
                                line['end_x'],
                                rtol=1e-8):
                            line['x'].append(p[0])

        points = []
        for line in lines:
            if line.get('start_x', None) is None:
                continue
            line['x'].sort()
            x1 = line['start_x']
            x2 = line['end_x']
            if line['x']:
                x = line['x'][0]  # first point
                line['x_dist'] = x - x1
                points.append((line['x_dist'], (x1+x)/2, line['y']))

                x = line['x'][-1]  # last point
                line['x_dist'] = x2 - x
                points.append((line['x_dist'], (x+x2)/2, line['y']))
            else:
                line['x_dist'] = x2 - x1  # no points between
                points.append((line['x_dist'], (x1+x2)/2, line['y']))

        points.sort()
        return (points[-1][1], points[-1][2])

    def get_point_inside(self, geom):
        """return point inside area"""
        mm = self.minmax()
        y = (mm[2]+mm[3])/2
        p1 = (mm[0]-5, y)
        p2 = (mm[1]+5, y)
        line = Line(Element(start=p1, end=p2))

        points = []
        for e in self.area:
            points += e.intersect_line(line, geom.rtol, geom.atol, True)

        if len(points) < 2:
            logger.debug("WARNING: get_point_inside() failed ({})".
                         format(len(points)))
            return None

        assert(len(points) > 1)

        my_points_sorted = [(p[0], p) for p in points]
        my_points_sorted.sort()
        my_p1 = my_points_sorted[0][1]   # Startpoint
        my_p2 = my_points_sorted[-1][1]  # Endpoint

        all_points_sorted = []
        for e in geom.elements(Shape):
            points = e.intersect_line(line, geom.rtol, geom.atol, True)
            for p in points:
                if greater(p[0], my_p1[0], rtol=1e-8):
                    if less(p[0], my_p2[0], rtol=1e-8):
                        all_points_sorted.append((p[0], p))

        if len(all_points_sorted) == 0:
            p_inside = ((my_p1[0]+my_p2[0])/2, y)
            if self.is_air():
                return self.get_best_point_inside(geom)
            return p_inside

        all_points_sorted.sort()
        all_p1 = all_points_sorted[0][1]
        all_p2 = all_points_sorted[-1][1]
        d1 = all_p1[0] - my_p1[0]
        d2 = my_p2[0] - all_p2[0]
        if d1 > d2:
            p_inside = ((my_p1[0]+all_p1[0])/2, y)
        else:
            p_inside = ((my_p2[0]+all_p2[0])/2, y)

        if self.is_air():
            return self.get_best_point_inside(geom)
        return p_inside

    def render(self, renderer, color='black', with_nodes=False, fill=True):
        if fill:
            if self.render_fill(renderer):
                color = 'black'

        for e in self.area:
            e.render(renderer, color, with_nodes)
        return

    def render_fill(self, renderer):
        color = self.color()
        if not color:
            return False
        alpha = self.color_alpha()

        if self.is_circle():
            e = self.area[0]
            renderer.fill_circle(e.center, e.radius, color, alpha)
        else:
            nodes = [n for n in self.virtual_nodes(render=True)]
            x = [n[0] for n in nodes]
            y = [n[1] for n in nodes]
            renderer.fill(x, y, color, alpha)
        return True

    def render_legend(self, renderer):
        return renderer.new_legend_handle(self.color(),
                                          self.color_alpha(),
                                          self.legend())

    def remove_edges(self, g, ndec):
        for e in self.area:
            try:
                g.remove_edge(e.node1(ndec), e.node2(ndec))
            except Exception:
                continue

    def is_circle(self):
        e = self.area[0]
        if len(self.area) == 1:
            return is_Circle(e)

        if isinstance(e, Arc):
            c = e.center
            r = e.radius
            a = 0.0
            for e in self.area:
                if not isinstance(e, Arc):
                    return False
                if not points_are_close(c, e.center):
                    return False
                if not np.isclose(r, e.radius):
                    return False
                a += e.get_angle_of_arc()
            return np.isclose(a, 2.0*np.pi)

        return False

    def is_half_circle(self, center, angle):
        for e in self.area:
            if isinstance(e, Line):
                if not np.isclose(angle, alpha_line(center, e.p1)):
                    return False
                if not np.isclose(angle, alpha_line(center, e.p2)):
                    return False
            elif isinstance(e, Arc):
                if not np.isclose(angle, alpha_line(center, e.center)):
                    return False
            else:
                return False
        return True

    def has_round_edges(self):
        arcs = 0
        for e in self.area:
            # if isinstance(e, Line):
            #    if not np.isclose(angle, alpha_line(center, e.p1)):
            #        return False
            #    if not np.isclose(angle, alpha_line(center, e.p2)):
            #        return False
            if isinstance(e, Arc):
                arcs += 1

        return arcs > 0

    def is_shaft_area(self, center):
        logger.debug("Begin of check shaft")

        if not self.is_touching_both_sides():
            logger.debug("End of check shaft: don't touch both sides")
            return False

        for n in self.list_of_nodes():
            a = alpha_line(center, n)
            if np.isclose(self.min_angle, a):
                continue
            if np.isclose(self.max_angle, a):
                continue
            d = distance(center, n)
            if np.isclose(d, self.min_dist, atol=0.05):
                continue
            if np.isclose(d, self.max_dist, atol=0.05):
                continue
            logger.debug("End of check shaft: no")
            return False
        logger.debug("End of check shaft: ok")
        return True

    def is_rectangle(self):
        lines = [[c, e.m(99999.0), e.length()]
                 for c, e in enumerate(self.area)
                 if isinstance(e, Line)]
        lines.sort()

        line_count = 1
        m_first = 0.0
        m_prev = 999.999999
        c_prev = -99
        m_all = []
        for c, m, l in lines:
            if c_prev >= 0:
                if np.isclose(m_prev, m, atol=0.001):
                    if c_prev+1 != c:
                        # Gleiche Steigung, aber keine Verlängerung
                        line_count += 1
                        m_all.append(m_prev)
                else:
                    line_count += 1
                    m_all.append(m_prev)
            else:
                m_first = m

            m_prev = m
            c_prev = c

        m_all.append(m_prev)

        if np.isclose(m_prev, m_first, atol=0.001):
            line_count -= 1

        if line_count == 4:
            logger.debug("is_rectangle: m={}".format(m_all))
            if not np.isclose(m_all[0], m_all[2], atol=0.001):
                return False
            if not np.isclose(m_all[1], m_all[3], atol=0.001):
                return False
            return True

        return False

    def is_mag_rectangle(self):
        lines_ceml = [[c, e, e.m(99999.0), e.length()]
                      for c, e in enumerate(self.area)]
        # c = Count
        # e = Element
        # m = Steigung
        # l = Länge
        # L = class Line
        if len(lines_ceml) < 4:
            return False

        logger.debug("=== BEGIN OF is_mag_rectangle() [{} lines]"
                     .format(len(lines_ceml)))

        c_prev = lines_ceml[0][0]
        a_prev = 999
        p = None

        e0 = lines_ceml[0][1]
        e0_p1 = e0.p1
        e0_p2 = e0.p2
        L_prev = isinstance(e0, Line)
        l_prev = lines_ceml[0][3]
        m_prev = lines_ceml[0][2]

        e1 = lines_ceml[1][1]
        e1_p1 = e1.p1
        e1_p2 = e1.p2

        if (points_are_close(e0_p2, e1_p1, atol=1e-02) or
                points_are_close(e0_p2, e1_p2, atol=1e-02)):
            a_prev = alpha_line(e0_p1, e0_p2)
            p = e0_p2
        elif (points_are_close(e0_p1, e1_p1, atol=1e-02) or
              points_are_close(e0_p1, e1_p2, atol=1e-02)):
            a_prev = alpha_line(e0_p2, e0_p1)
            p = e0_p1
        else:
            logger.error(
                "ERROR: is_mag_rectangle(): points are not close together")
            logger.error("       e0 p1={}, p2={}".format(e0_p1, e0_p2))
            logger.error("       e1 p1={}, p2={}".format(e1_p1, e1_p2))
            return False

        def alpha_current(p, e):
            if points_are_close(p, e.p1, atol=1e-02):
                return e.p2, alpha_line(e.p1, e.p2), isinstance(e, Line)
            if points_are_close(p, e.p2, atol=1e-02):
                return e.p1, alpha_line(e.p2, e.p1), isinstance(e, Line)
            logger.error(
                "ERROR: is_mag_rectangle(): points are not close together")
            logger.error("       p={}, p1={}, p2={}".format(p, e.p1, e.p2))
            return None, None, False

        lines_clamL = []
        for c, e, m, l in lines_ceml[1:]:
            p, a_curr, L_curr = alpha_current(p, e)
            if not p:
                return False

            if is_same_angle(a_prev, a_curr, atol=0.01):
                # its the same angle and both are Lines
                # assert(np.isclose(m_prev, m, atol=0.001))
                if c_prev+1 != c:
                    logger.debug(" - ok, but not an extension")
                    # ..., but not an extension
                    lines_clamL.append([c_prev, l_prev, a_prev, m])
                    l_prev = e.length()
                else:
                    # ... and an extension
                    l_prev += e.length()
                    logger.debug(" - ok, it's an extension")
            else:
                # it's a different angle
                logger.debug(" - diff, angle {} and {} not equal "
                             .format(a_prev, a_curr))
                lines_clamL.append([c_prev, l_prev, a_prev, m_prev, L_prev])
                l_prev = e.length()

            a_prev = a_curr
            L_prev = L_curr
            m_prev = m
            c_prev = c

        lines_clamL.append([c_prev, l_prev, a_prev, m_prev, L_prev])
        if np.isclose(lines_clamL[0][2], lines_clamL[-1][2], atol=0.001):
            # Gleicher Winkel am Anfang und am Ende
            lines_clamL[0][1] += lines_clamL[-1][1]  # length
            del lines_clamL[-1]
            logger.debug(" > last entry deleted")

        if len(lines_clamL) < 4:
            logger.debug("=== END OF is_mag_rectangle(): NO RECTANGLE #1")
            return False

        lines_lmcL = [[l, m, c, L] for c, l, a, m, L in lines_clamL]
        lines_lmcL.sort(reverse=True)

        if not np.isclose(lines_lmcL[0][1], lines_lmcL[1][1], atol=0.05):
            # Die Steigungen der zwei längsten Linien müssen gleich sein
            logger.debug("--- m %s <> %s ---",
                         lines_lmcL[0][1],
                         lines_lmcL[1][1])
            logger.debug("--- l %s, %s, %s ---",
                         lines_lmcL[0][0],
                         lines_lmcL[1][0],
                         lines_lmcL[2][0])
            logger.debug("=== END OF is_mag_rectangle(): NO RECTANGLE #2")
            return False

        def excursion_to_same_direction(clam):
            if len(clam) < 4:
                return False

            alpha = alpha_angle(clam[0][2], clam[1][2])
            clockwise = not alpha < np.pi

            angle_prev = clam[1][2]
            for c, l, angle_curr, m, t in clam[2:]:
                alpha = alpha_angle(angle_prev, angle_curr)
                if clockwise:
                    if alpha < np.pi:
                        return False
                else:
                    if alpha > np.pi:
                        return False
                angle_prev = angle_curr
            return True  # end of all_lines_with_same_direction()

        lines_cmL = [[c, m, L] for l, m, c, L in lines_lmcL[0:4]]
        lines_cmL.sort()

        if np.isclose(lines_cmL[0][1], lines_cmL[2][1], atol=0.001):
            if not (lines_cmL[0][2] and lines_cmL[2][2]):
                logger.debug("=== END OF is_mag_rectangle(): not 2 lines #1")
                return False
            ok = excursion_to_same_direction(lines_clamL)
            logger.debug("=== END OF is_mag_rectangle(): OK = {} #1"
                         .format(ok))
            return ok
        if np.isclose(lines_cmL[1][1], lines_cmL[3][1], atol=0.001):
            if not (lines_cmL[1][2] and lines_cmL[3][2]):
                logger.debug("=== END OF is_mag_rectangle(): not 2 lines #2")
                return False

            ok = excursion_to_same_direction(lines_clamL)
            logger.debug("=== END OF is_mag_rectangle(): OK = {} #2"
                         .format(ok))
            return ok

        logger.debug("=== END OF is_mag_rectangle(): NO RECTANGLE #3")
        return False

    def get_mag_orient_rectangle(self):
        lines = [[e.m(99999.0), e.length(), alpha_line(e.p1, e.p2)]
                 for e in self.area
                 if isinstance(e, Line)]
        lines.sort()

        m_prev = 999.999999
        a_prev = 0.0
        l_total = 0.0
        line_length = []
        for m, l, a in lines:
            if np.isclose(m_prev, m):
                l_total += l
            else:
                if l_total > 0.0:
                    line_length.append((l_total, m_prev, a_prev))
                l_total = l
                m_prev = m
                a_prev = a

        if l_total > 0.0:
            line_length.append((l_total, m_prev, a_prev))
        line_length.sort(reverse=True)

        alpha = line_length[0][2]
        if alpha < 0.0:
            alpha += np.pi
        alpha = alpha + np.pi/2
        if alpha > np.pi:
            alpha = alpha - np.pi
        return alpha

    def get_mag_orientation(self):
        if self.mag_rectangle:
            return self.get_mag_orient_rectangle()

        if self.close_to_endangle:
            if self.close_to_startangle:
                return middle_angle(self.min_angle, self.max_angle)
            else:
                return self.max_angle
        else:
            return middle_angle(self.min_angle, self.max_angle)

    def around_windings(self, areas):
        for a in areas:
            if a.is_winding():
                if not self.is_identical(a):
                    if self.is_inside(a):
                        return True
                    elif self.is_touching(a):
                        return True
        return False

    def is_touching_areas(self, areas):
        for a in areas:
            if self.is_touching(a):
                return True
        return False

    def has_iron_separator(self):
        for e in self.area:
            if e.has_attribute('iron_sep'):
                return True
        return False

    def mark_stator_subregions(self,
                               is_inner,
                               stator_size,
                               mirrored,
                               alpha,
                               center,
                               r_in,
                               r_out):
        alpha = round(alpha, 6)

        if self.is_circle():
            self.type = 0  # air
            return self.type

        ag_delta = (r_out - r_in) / 500.0
        if is_inner:
            self.close_to_ag = greater_equal(self.max_dist + ag_delta, r_out)
            close_to_opposition = np.isclose(r_in, self.min_dist)
            airgap_radius = r_out
            opposite_radius = r_in
            airgap_toleranz = -(self.max_dist - self.min_dist) / 50.0  # 2%
        else:
            self.close_to_ag = less_equal(self.min_dist - ag_delta, r_in)
            close_to_opposition = np.isclose(r_out, self.max_dist)
            airgap_radius = r_in
            opposite_radius = r_out
            airgap_toleranz = (self.max_dist - self.min_dist) / 50.0  # 2%

        self.close_to_startangle = np.isclose(self.min_angle, 0.0,
                                              1e-04, 1e-04)
        self.close_to_endangle = np.isclose(self.max_angle, alpha,
                                            1e-04, 1e-04)
        self.surface = self.area_size()

        logger.debug("\n***** mark_stator_subregions [{}] *****"
                     .format(self.id))
        logger.debug(" - close_to_ag        : %s", self.close_to_ag)
        logger.debug(" - close_to_opposition: %s", close_to_opposition)
        logger.debug(" - airgap_radius      : %3.12f", airgap_radius)
        logger.debug(" - airgap_toleranz    : %3.12f", airgap_toleranz)
        logger.debug(" - opposite radius    : %3.12f", opposite_radius)
        logger.debug(" - close_to_startangle: %s", self.close_to_startangle)
        logger.debug(" - close_to_endangle  : %s", self.close_to_endangle)
        logger.debug(" - alpha              : %3.12f", alpha)
        logger.debug(" - min_angle          : %3.12f", self.min_angle)
        logger.debug(" - max_angle          : %3.12f", self.max_angle)
        logger.debug(" - min_dist           : %3.12f", self.min_dist)
        logger.debug(" - max_dist           : %3.12f", self.max_dist)
        logger.debug(" - surface size       : %3.12f", self.surface)

        if is_inner:
            # looking for shaft
            if close_to_opposition and not self.close_to_ag:
                if self.is_shaft_area(center):
                    self.type = 10  # shaft
                    logger.debug("***** shaft (close to opposition)\n")
                    return self.type

        if close_to_opposition:
            self.type = 5  # iron yoke (Joch)
            logger.debug("***** iron yoke #1\n")
            return self.type

        if self.close_to_startangle and self.close_to_endangle:
            self.type = 5  # iron yoke (Joch)
            logger.debug("***** iron yoke #2\n")
            return self.type

        if self.close_to_ag:  # close to airgap
            mm = self.minmax_angle_dist_from_center(center,
                                                    airgap_radius +
                                                    airgap_toleranz)
            self.min_air_angle = mm[0]
            self.max_air_angle = mm[1]
            air_alpha = round(alpha_angle(mm[0], mm[1]), 3)
            logger.debug(" - min_air_alpha      : {}".format(mm[0]))
            logger.debug(" - max_air_alpha      : {}".format(mm[1]))
            logger.debug(" - air_alpha          : {}".format(air_alpha))

            if self.alpha / air_alpha > 2:
                logger.debug("***** windings near airgap\n")
                self.type = 2  # windings
            else:
                self.type = 9  # air or iron near windings and near airgap?
                logger.debug("***** air or iron ??\n")
            return self.type

        if self.close_to_startangle:
            if self.is_half_circle(center, self.min_angle):
                self.type = 0  # air
                logger.debug("***** air (part of a circle)\n")
                return self.type

        if self.close_to_endangle:
            if self.is_half_circle(center, self.max_angle):
                self.type = 0  # air
                logger.debug("***** air (part of a circle)\n")
                return self.type

        def bad_winding_position():
            if is_inner:
                radius_third = airgap_radius - (airgap_radius - opposite_radius) * 0.33
                if self.max_dist < radius_third:
                    return True
            else:  # outer
                radius_third = airgap_radius + (opposite_radius - airgap_radius) * 0.33
                if self.min_dist > radius_third:
                    return True
            return False

        if self.min_angle > 0.001:
            if self.max_angle < alpha - 0.001:
                if bad_winding_position():
                    self.type = 12  # windings or air
                    logger.debug("***** windings or air #1\n")
                else:
                    self.type = 2  # windings
                    logger.debug("***** windings #1\n")
                return self.type
            if mirrored:
                if bad_winding_position():
                    self.type = 12  # windings or air
                    logger.debug("***** windings or air #2\n")
                else:
                    self.type = 2  # windings
                    logger.debug("***** windings #2\n")
                return self.type

            self.type = 0  # air
            logger.debug("***** air #3")

        if self.close_to_startangle or self.close_to_endangle:
            f = self.surface / stator_size
            if f < 0.02:  # area_size less then 2 percent of stator size
                # Luftloch
                self.type = 0  # air
                logger.debug("***** small area => air\n")
            else:
                self.type = 9  # air or iron near windings and near airgap?
                logger.debug("***** air or iron close to border\n")
            return self.type

        logger.debug("***** air #4\n")
        return 0

    def mark_rotor_subregions(self, is_inner, mirrored, alpha,
                              center, r_in, r_out):
        logger.debug("mark_rotor_subregions")

        alpha = round(alpha, 6)

        if self.is_circle():
            self.type = 0  # air
            logger.debug(">>> air is a circle")
            return self.type

        if is_inner:
            self.close_to_ag = np.isclose(r_out, self.max_dist, atol=0.005)
            close_to_opposition = greater_equal(r_in * 1.05, self.min_dist)
            airgap_radius = r_out
            opposite_radius = r_in
            airgap_toleranz = -(self.max_dist - self.min_dist) / 50.0  # 2%
        else:
            self.close_to_ag = np.isclose(r_in, self.min_dist, atol=0.005)
            close_to_opposition = greater_equal(self.max_dist * 1.05, r_out)
            airgap_radius = r_in
            opposite_radius = r_out
            airgap_toleranz = (self.max_dist - self.min_dist) / 50.0  # 2%

        self.close_to_startangle = np.isclose(self.min_angle, 0.0,
                                              1e-04, 1e-04)
        self.close_to_endangle = np.isclose(self.max_angle, alpha,
                                            1e-04, 1e-04)

        logger.debug("\n***** mark_rotor_subregions [{}] *****"
                     .format(self.id))
        logger.debug(" - close_to_ag        : %s", self.close_to_ag)
        logger.debug(" - close_to_opposition: %s", close_to_opposition)
        logger.debug(" - min dist           : %3.12f", self.min_dist)
        logger.debug(" - max dist           : %3.12f", self.max_dist)
        logger.debug(" - airgap radius      : %3.12f", airgap_radius)
        logger.debug(" - opposite radius    : %3.12f", opposite_radius)
        logger.debug(" - close_to_startangle: %s", self.close_to_startangle)
        logger.debug(" - close_to_endangle  : %s", self.close_to_endangle)
        logger.debug(" - alpha              : %3.12f", alpha)
        logger.debug(" - min_angle          : %3.12f", self.min_angle)
        logger.debug(" - max_angle          : %3.12f", self.max_angle)

        if self.has_iron_separator():
            logger.debug("***** iron (has iron separator)\n")
            self.type = 1  # iron
            return self.type

        if is_inner:
            # looking for shaft
            if close_to_opposition and not self.close_to_ag:
                if self.is_shaft_area(center):
                    self.type = 10  # shaft
                    logger.debug("***** shaft (close to opposition)\n")
                    return self.type

        if close_to_opposition:
            self.type = 1  # iron
            logger.debug("***** iron (close to opposition)\n")
            return self.type

        if self.close_to_startangle and self.close_to_endangle:
            self.type = 1  # iron
            logger.debug("***** iron (close to both sides)\n")
            return self.type

        self.mag_rectangle = self.is_mag_rectangle()

        if self.close_to_ag:
            mm = self.minmax_angle_dist_from_center(center,
                                                    airgap_radius +
                                                    airgap_toleranz)
            air_alpha = round(alpha_angle(mm[0], mm[1]), 3)
            logger.debug(" - air_alpha          : {}".format(air_alpha))

            if air_alpha / alpha < 0.2:
                self.phi = self.get_mag_orientation()
                self.type = 8  # air or magnet ?
                logger.debug("***** air #1 (close to airgap)\n")
                return self.type

            if air_alpha / alpha > 0.6:
                self.phi = self.get_mag_orientation()
                self.type = 3  # magnet
                logger.debug("***** magnet (close to airgap)\n")
            else:
                self.phi = self.get_mag_orientation()
                self.type = 9  # iron or magnet ?
                logger.debug("***** iron or magnet(close to airgap)\n")
            return self.type

        if self.mag_rectangle:
            self.phi = self.get_mag_orientation()
            self.type = 4  # magnet embedded
            logger.debug("***** magnet (embedded, phi={})\n".format(
                self.phi))
            return self.type

        if not (self.close_to_startangle or self.close_to_endangle):
            self.type = 0  # air
            logger.debug("***** air (somewhere)\n")
            return self.type

        if self.close_to_startangle:
            if self.is_half_circle(center, self.min_angle):
                self.type = 0  # air
                logger.debug("***** air (part of a circle)\n")
                return self.type

        if self.close_to_endangle:
            if self.is_half_circle(center, self.max_angle):
                self.type = 0  # air
                logger.debug("***** air (part of a circle)\n")
                return self.type

        self.type = 0  # air
        logger.debug("***** air (remains)\n")
        return self.type

    def mark_unknown_subregions(self, mirrored, alpha,
                                center, r_in, r_out):
        logger.debug("mark_unknown_subregions")

        if self.is_circle():
            self.type = 0  # air
            logger.debug(">>> air is a circle")
            return self.type

        self.close_to_startangle = np.isclose(self.min_angle, 0.0)
        self.close_to_endangle = np.isclose(self.max_angle, alpha)

        if self.is_mag_rectangle():
            self.type = 4  # magnet embedded
            logger.debug(">>> magnet embedded")
            self.phi = self.get_mag_orient_rectangle()
            return self.type

        close_to_max_radius = np.isclose(r_out, self.max_dist)
        close_to_min_radius = np.isclose(r_in, self.min_dist)

        if close_to_max_radius and close_to_min_radius:
            self.type = 1  # iron
            logger.debug(">>> iron close to min- and max-radius")
            return self.type

        if self.close_to_startangle and self.close_to_endangle:
            self.type = 1  # iron
            logger.debug(">>> iron close to start- and end-angle")
            return self.type

        self.type = 0  # air
        logger.debug(">>> air remains")
        return self.type

    def mark_airgap_corners(self, start_cp, end_cp):
        for n in self.list_of_nodes():
            if self.close_to_startangle:
                if points_are_close(n, start_cp):
                    self.close_to_ag_startcorner = True
            if self.close_to_endangle:
                if points_are_close(n, end_cp):
                    self.close_to_ag_endcorner = True

    def area_size(self):
        nodes = [n for n in self.list_of_nodes()]
        return area_size(nodes)

    def set_surface(self, mirrored):
        self.surface = self.area_size()
        if self.close_to_endangle and mirrored:
            self.surface = self.area_size() * 2.0
        else:
            self.surface = self.area_size()

    def print_area(self):
        center = [0.0, 0.0]
        for s in self.area:
            mm = s.minmax_angle_from_center(center)
            print(" --- angle min={}, max={}".format(mm[0], mm[1]))

    def __lt__(self, a):
        if self.symmetry != a.symmetry:
            return self.symmetry > a.symmetry

        if self.sym_type != a.sym_type:
            return self.sym_type > a.sym_type

        if self.count != a.count:
            return self.count > a.count

        if not np.isclose(self.height, a.height):
            return self.height > a.height

        if self.sym_tolerance > 0.0:
            if not np.isclose(round(self.min_dist, 4),
                              round(a.min_dist, 4), 1e-03,
                              self.sym_tolerance):
                return less_equal(self.min_dist, a.min_dist)
            if not np.isclose(round(self.max_dist, 4),
                              round(a.max_dist, 4), 1e-03,
                              self.sym_tolerance):
                return less_equal(self.max_dist, a.max_dist)
            if not np.isclose(round(self.alpha, 2),
                              round(a.alpha, 2), 1e-01, 1e-01):
                return less_equal(self.alpha, a.alpha)
        else:
            if not np.isclose(round(self.min_dist, 2),
                              round(a.min_dist, 2)):
                return less_equal(self.min_dist, a.min_dist)
            if not np.isclose(round(self.max_dist, 2),
                              round(a.max_dist, 2)):
                return less_equal(self.max_dist, a.max_dist)
            if not np.isclose(round(self.alpha, 2),
                              round(a.alpha, 2), 1e-01, 1e-02):
                return less_equal(self.alpha, a.alpha)

        return self.min_angle < a.min_angle

    def nested_areas_inside(self):
        for id, a in self.areas_inside.items():
            yield id
            for i in a.nested_areas_inside():
                yield i

    def list_of_nested_areas_inside(self):
        for id, a in self.areas_inside.items():
            for i in a.nested_areas_inside():
                yield i

    def crunch_area(self, geom):
        n1_prev = None
        e_prev = None
        logger.debug("crunch area %s", self.identifier())

        if self.is_circle():
            return 0

        c = 0
        for n1, n2, e in self.list_of_elements():
            if e_prev is not None:
                if len([nbr for nbr in geom.g.neighbors(n1)]) == 2:
                    e_new = e_prev.concatenate(n1_prev, n2, e)
                    if e_new is not None:
                        e_prev_dict = geom.g.get_edge_data(n1_prev, n1)
                        e_dict = geom.g.get_edge_data(n1, n2)

                        logger.debug("--> remove from %s to %s [%s, %s, %s]",
                                     n1_prev,
                                     n1,
                                     e_prev_dict[0],
                                     e_prev_dict[1],
                                     e_prev_dict[2])
                        logger.debug("    remove %s", e_prev)
                        geom.remove_edge(e_prev)

                        logger.debug("--> remove from %s to %s [%s, %s, %s]",
                                     n1,
                                     n2,
                                     e_dict[0],
                                     e_dict[1],
                                     e_dict[2])
                        logger.debug("    remove %s", e)
                        if e.get_node_number(n1) == 1:
                            flag1 = e_dict[1]
                            flag2 = e_dict[2]
                        else:
                            flag1 = e_dict[2]
                            flag2 = e_dict[1]
                        geom.remove_edge(e)

                        logger.debug("--> add from %s to %s",
                                     n1_prev,
                                     n2)
                        logger.debug("    add %s", e_new)
                        geom.add_edge(n1_prev, n2, e_new)

                        e_new_dict = geom.g.get_edge_data(n1_prev, n2)
                        e_new_dict[0] = True
                        if e_new.get_node_number(n1_prev) == 1:
                            e_new_dict[1] = flag1
                            e_new_dict[2] = flag2
                        else:
                            e_new_dict[1] = flag2
                            e_new_dict[2] = flag1

                        logger.debug("    new dict: [%s, %s, %s]",
                                     e_new_dict[0],
                                     e_new_dict[1],
                                     e_new_dict[2])
                        e_prev = e_new
                        c += 1
                        continue

            n1_prev = n1
            e_prev = e
        return c

    def __str__(self):
        return "Area {}\n".format(self.id) + \
            "distance...............: from {} to {}\n".\
            format(round(self.min_dist, 4), round(self.max_dist, 4)) + \
            "height.................: {}\n".format(self.height) + \
            "alpha..................: {}\n".format(self.alpha) + \
            "angle..................: from {} to {}\n".\
            format(round(self.min_angle, 6), round(self.max_angle, 6)) + \
            "delta..................: {}\n".format(self.delta) + \
            "number.................: {}\n".format(self.count) + \
            "equal areas............: {}\n".format(len(self.equal_areas)) + \
            "symmetry...............: {}\n".format(self.symmetry) + \
            "symmetry type..........: {}\n".format(self.sym_type) + \
            "close to airgap........: {}\n".format(self.close_to_ag) + \
            "close to startangle....: {}\n".format(self.close_to_startangle) + \
            "close to endangle......: {}\n".format(self.close_to_endangle) + \
            "close to ag startcorner: {}\n".format(self.close_to_ag_startcorner) + \
            "close to ag endcorner..: {}\n".format(self.close_to_ag_endcorner)



"""
    femagtools.tks
    ~~~~~~~~~~~~~~

    Manage TKS magnetizing curve data files



"""
import sys
import numpy as np
import os
import re
import codecs
import femagtools.losscoeffs as lc
import femagtools.mcv
import json
import logging

logger = logging.getLogger(__name__)

MUE0 = 4e-7*np.pi  # 1.2566371E-06
fo = 50.
Bo = 1.5
numPattern = re.compile(r'([+-]?\d+(?:\.\d+)?(?:[eE][+-]\d+)?)')
HBpattern = re.compile(r'H.+\s+B')
BPpattern = re.compile(r'B.+\s+P')


def readlist(section):
    x = []
    for l in section:
        if not l:
            break
        vals = [float(n) for n in numPattern.findall(l.replace(',', '.'))]
        if vals:
            x.append(vals)
    return list(zip(*x))


class Reader(object):

    def __init__(self, filename, filecontent=None):
        self.version_mc_curve = 0
        self.mc1_type = femagtools.mcv.MAGCRV

        self.curve = []
        self.name = os.path.splitext(
            os.path.basename(filename))[0]
        self.mc1_title = 'ThyssenKrupp Steel'
        self.curve = [{}]
        self.curve[0]['hi'] = []
        self.curve[0]['bi'] = []
        self.fo = fo
        self.Bo = Bo
        self.ch = 0.0
        self.ch_freq = 0.0
        self.cw = None
        self.cw_freq = None
        self.b_coeff = None
        self.rho = 7.6
        self.losses = dict(f=[], B=[])
        pfe = []

        # filecontent is used
        if filecontent:
            content = [l.strip() for l in filecontent.split('\n')]
        else:
            # only filename
            with codecs.open(filename, encoding='utf-8', errors='ignore') as f:
                content = [l.strip() for l in f.readlines()]
        if content:
            for i, l in enumerate(content):
                if HBpattern.match(l):
                    hbj = readlist(content[i+1:])
                    self.curve[0]['hi'] = hbj[0]
                    self.curve[0]['bi'] = hbj[1]

                elif l.startswith('Material Name'):
                    self.name = l.split(':')[1].strip()

                elif l.startswith('Comment'):
                    self.mc1_title = l.split(':')[1].strip()

                elif l.startswith('Mass Density'):
                    d = numPattern.findall(l.replace(',', '.'))
                    self.rho = float(d[0])
                    if l.split()[-1] == 'kg/m^3':
                        self.rho /= 1e3

                elif l.startswith('f='):
                    fref = numPattern.findall(l.replace(',', '.'))
                    fxref = float(fref[0])

                elif BPpattern.match(l):
                    b, p = readlist(content[i+1:])
                    self.losses['f'].append(fxref)
                    self.losses['B'].append(b)
                    pfe.append(p)
        logger.info("%s Bmax %3.2f", filename, max(self.curve[0]['bi']))

        if pfe and not np.isscalar(self.losses['B'][0]):
            colsize = max([len(p) for p in pfe])
            losses = [list(p) + [0]*(colsize-len(p)) for p in pfe]
            z = lc.fitjordan(
                self.losses['f'],
                self.losses['B'],
                losses,
                self.Bo,
                self.fo)
            logger.info("Jordan loss coeffs %s", z)
            self.ch = z[0]
            self.ch_freq = z[1]
            self.cw = z[2]
            self.cw_freq = z[3]
            self.b_coeff = z[4]

            z = lc.fitsteinmetz(
                self.losses['f'],
                self.losses['B'],
                losses,
                self.Bo,
                self.fo)
            logger.info("Steinmetz loss coeffs %s", z)

            self.losses['cw'] = z[0]
            self.losses['cw_freq'] = z[1]
            self.losses['b_coeff'] = z[2]

            self.losses['Bo'] = self.Bo
            self.losses['fo'] = self.fo

            # must normalize pfe matrix:
            (self.losses['B'],
             self.losses['pfe']) = femagtools.mcv.norm_pfe(self.losses['B'], pfe)

    def __getitem__(self, index):
        return self.__getattribute__(index)

    def getValues(self):
        """return values as mcv dict"""
        return {
            'name': self.name,
            'desc': self.mc1_title,
            'cversion': self.version_mc_curve,
            'ctype': self.mc1_type,
            'curve': self.curve,
            'fo': self.fo,
            'Bo': self.Bo,
            'ch': self.ch,
            'ch_freq': self.ch_freq,
            'cw': self.cw,
            'cw_freq': self.cw_freq,
            'b_coeff': self.b_coeff,
            'rho': self.rho,
            'losses': self.losses}


def read(filename, filecontent=None):
    """read Thyssen File TKS and return mc dict"""
    tks = Reader(filename, filecontent=filecontent)
    return tks.getValues()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s')
    if len(sys.argv) == 2:
        filename = sys.argv[1]
    else:
        filename = sys.stdin.readline().strip()

    tks = Reader(filename)
    if tks.losses:
        import matplotlib.pylab as pl
        import femagtools.plot
        import numpy as np
        cw = tks.cw
        beta = tks.cw_freq
        ch = tks.ch
        alpha = tks.ch_freq
        gamma = tks.b_coeff

        femagtools.plot.felosses(tks.losses,
                                 (ch, alpha, cw, beta, gamma),
                                 #                                 (tks.losses['cw'],
                                 #                                  tks.losses['cw_freq'],
                                 #                                  tks.losses['b_coeff']),
                                 title=filename, log=False)
        pl.show()

    mcv = tks.getValues()
    json.dump(mcv, sys.stdout)



# -*- coding: utf-8 -*-
"""
    femagtools.nc
    ~~~~~~~~~~~~~

    Read FEMAG netcdf model files
"""
import logging
import warnings
import netCDF4
from femagtools import isa7

logger = logging.getLogger('femagtools.nc')


class Reader(object):
    """
    Open and Read NetCDF file

    Arguments:
        filename: name of NetCDF nc file to be read
    """

    def __init__(self, filename):
        ds = netCDF4.Dataset(filename)
        self.POINT_ISA_POINT_REC_PT_CO_X = []
        self.POINT_ISA_POINT_REC_PT_CO_Y = []
        self.LINE_ISA_LINE_REC_LN_PNT_1 = []
        self.LINE_ISA_LINE_REC_LN_PNT_2 = []
        try:
            grp = ds.groups['points']
            (self.POINT_ISA_POINT_REC_PT_CO_X,
             self.POINT_ISA_POINT_REC_PT_CO_Y) = [
                 grp.variables[k][:] for k in ('co_x', 'co_y')]
        except KeyError:
            pass
        try:
            grp = ds.groups['lines']
            (self.LINE_ISA_LINE_REC_LN_PNT_1,
             self.LINE_ISA_LINE_REC_LN_PNT_2) = [
                 grp.variables[k][:] for k in ('pnt_1', 'pnt_2')]
        except KeyError:
            pass
        grp = ds.groups['nodes']
        (self.NODE_ISA_NOD_EL_PNTR,
         self.NODE_ISA_NODE_REC_ND_BND_CND,
         self.NODE_ISA_NODE_REC_ND_PER_NOD,
         self.NODE_ISA_NODE_REC_ND_CO_1,
         self.NODE_ISA_NODE_REC_ND_CO_2,
         self.NODE_ISA_ND_CO_RAD,
         self.NODE_ISA_ND_CO_PHI,
         self.NODE_ISA_NODE_REC_ND_VP_RE,
         self.NODE_ISA_NODE_REC_ND_VP_IM) = [
             grp.variables[k][:-1]
             for k in ('bnd_cnd', 'bnd_cnd', 'per_nod',
                       'co_1', 'co_2', 'co_rad', 'co_phi', 'vp_re', 'vp_im')]
        logger.debug('Nodes: %d', len(self.NODE_ISA_NODE_REC_ND_CO_1))

        grp = ds.groups['node_elements']
        (self.NODE_ELE_ISA_NOD_EL_KEY,
         self.NODE_ELE_ISA_NOD_NXT_EL_PNTR) = [
             grp.variables[k][:-1]
             for k in ('el_key', 'nxt_el_pntr')]

        grp = ds.groups['nodechains']
        (self.NDCHN_ISA_NDCHN_REC_NC_NOD_1,
         self.NDCHN_ISA_NDCHN_REC_NC_NOD_2,
         self.NDCHN_ISA_NDCHN_REC_NC_NOD_MID) = [
             grp.variables[k][:-1]
             for k in ('nod_1', 'nod_2', 'nod_mid')]

        grp = ds.groups['elements']
        (self.ELEM_ISA_EL_NOD_PNTR,
         self.ELEM_ISA_ELEM_REC_EL_TYP,
         self.ELEM_ISA_ELEM_REC_EL_SE_KEY,
         self.ELEM_ISA_ELEM_REC_EL_RELUC,
         self.ELEM_ISA_ELEM_REC_EL_RELUC_2,
         self.ELEM_ISA_ELEM_REC_EL_MAG_1,
         self.ELEM_ISA_ELEM_REC_EL_MAG_2,
         self.ELEM_ISA_ELEM_REC_LOSS_DENS) = [
             grp.variables[k][:-1]
             for k in ('nod_pntr', 'type', 'se_key', 'reluc', 'reluc_2',
                       'mag_1', 'mag_2', 'loss_dens')]
        try:
            self.ELEM_ISA_ELEM_REC_TEMPERATURE = grp.variables['temperature'][:-1]
        except:
            pass
        logger.debug('Elements: %d', len(self.ELEM_ISA_ELEM_REC_EL_TYP))

        grp = ds.groups['element_nodes']
        (self.ELE_NOD_ISA_ND_KEY,
         self.ELE_NOD_ISA_NXT_ND_PNTR) = [
             grp.variables[k][:-1]
             for k in ('nd_key', 'nxt_nd_pntr')]

        grp = ds.groups['superelements']
        (self.SUPEL_ISA_SE_NDCHN_PNTR,
         self.SUPEL_ISA_SE_EL_PNTR,
         self.SUPEL_ISA_SUPEL_REC_SE_COL,
         self.SUPEL_ISA_SUPEL_REC_SE_MCV_TYP,
         self.SUPEL_ISA_SUPEL_REC_SE_COND_TYP,
         self.SUPEL_ISA_SUPEL_REC_SE_VEL_SYS,
         self.SUPEL_ISA_SUPEL_REC_SE_SR_KEY,
         self.SUPEL_ISA_SUPEL_REC_SE_VELO_1,
         self.SUPEL_ISA_SUPEL_REC_SE_VELO_2,
         self.SUPEL_ISA_SUPEL_REC_SE_CONDUC,
         self.SUPEL_ISA_SUPEL_REC_SE_LENGHT,
         self.SUPEL_ISA_SUPEL_REC_SE_CURD_RE,
         self.SUPEL_ISA_SUPEL_REC_SE_CURD_IM,
         self.SUPEL_ISA_SUPEL_REC_SE_FILLFACTOR) = [
            grp.variables[k][:]
            for k in ('ndch_pntr', 'el_pntr', 'color', 'mcv_type',
                      'cond_type', 'vel_sys', 'sr_key', 'velo_1',
                      'velo_2', 'conduc', 'length', 'curd_re', 'curd_im',
                      'fillfactor')]
        try:
            self.SUPEL_ISA_SUPEL_REC_SE_TEMP_COEF = grp.variables['temp_coef'][:]
        except:
            pass

        grp = ds.groups['superelement_nodechains']
        (self.SE_NDCHN_ISA_NC_KEY,
         self.SE_NDCHN_ISA_NXT_NC_PNTR) = [
            grp.variables[k][:]
            for k in ('nc_key', 'nxt_nc_pntr')]

        grp = ds.groups['superelement_elements']
        (self.SE_EL_ISA_EL_KEY,
         self.SE_EL_ISA_NXT_EL_PNTR) = [
            grp.variables[k][:]
            for k in ('el_key', 'nxt_el_pntr')]

        grp = ds.groups['subregions']
        (self.SR_ISA_SR_SE_PNTR,
         self.SR_ISA_SR_REC_SR_TYP,
         self.SR_ISA_SR_REC_SR_COL,
         names,
         self.SR_ISA_SR_REC_SR_CUR_DIR,
         self.SR_ISA_SR_REC_SR_WB_KEY,
         self.SR_ISA_SR_REC_SR_NTURNS,
         self.SR_ISA_SR_REC_SR_SV_PNTR,
         self.SR_ISA_SR_REC_SR_ARRAY,
         self.SR_ISA_SR_REC_SR_GCUR_RE,
         self.SR_ISA_SR_REC_SR_GCUR_IM,
         self.SR_ISA_SR_REC_SR_VOLT_RE,
         self.SR_ISA_SR_REC_SR_VOLT_IM) = [
            grp.variables[k][:]
            for k in ('se_pntr', 'type', 'color', 'name', 'cur_dir',
                      'wb_key', 'nturns', 'sv_pntr', 'array',
                      'gcur_re', 'gcur_im', 'volt_re', 'volt_im')]
        self.SR_ISA_SR_REC_SR_NAME = [''.join(str(n, encoding='utf-8'))
                                      for n in names]
        grp = ds.groups['subregion_superelements']
        (self.SR_SE_ISA_SE_KEY,
         self.SR_SE_ISA_NXT_SE_PNTR) = [
            grp.variables[k][:]
            for k in ('se_key', 'nxt_se_pntr')]
        try:
            grp = ds.groups['windings']
            (self.WB_ISA_WB_SR_PNTR,
             self.WB_ISA_WB_REC_WB_COL,
             names,
             self.WB_ISA_WB_REC_WB_TURN,
             self.WB_ISA_WB_REC_WB_SR_NUM,
             self.WB_ISA_WB_REC_WB_WND_KEY,
             self.WB_ISA_WB_REC_WB_GCUR_RE,
             self.WB_ISA_WB_REC_WB_GCUR_IM,
             self.WB_ISA_WB_REC_WB_VOLT_RE,
             self.WB_ISA_WB_REC_WB_VOLT_IM,
             self.WB_ISA_WB_REC_WB_IMPDZ_RE,
             self.WB_ISA_WB_REC_WB_IMPDZ_IM) = [
                 grp.variables[k][:]
                 for k in ('sr_pntr', 'color', 'name', 'turn', 'sr_num',
                           'wnd_key', 'gcur_re', 'gcur_im',
                           'volt_re', 'volt_im', 'impdz_re', 'impdz_im')]
            self.WB_ISA_WB_REC_WB_UNIT_RES = self.WB_ISA_WB_REC_WB_TURN
            self.WB_ISA_WB_REC_WB_NAME = [''.join(str(n, encoding='utf-8'))
                                          for n in names]
        except:
            pass
        try:
            grp = ds.groups['winding_subregions']
            (self.WB_SR_ISA_SR_KEY,
             self.WB_SR_ISA_NXT_SR_PNTR) = [
                 grp.variables[k][:]
                 for k in ('sr_key', 'nxt_sr_pntr')]
        except:
            pass
        try:
            grp = ds.groups['machine']
            self.speed = float(grp.variables['speed'].getValue().data)
            self.FC_RADIUS = float(grp.variables['fc_radius'].getValue().data)
            self.pole_pairs = int(grp.variables['pole_pairs'].getValue().data)
            self.delta_node_angle = float(  # rad
                grp.variables['delta_node_angle'].getValue().data)
            self.poles_sim = int(grp.variables['poles_sim'].getValue().data)
            self.slots = int(grp.variables['num_slots'].getValue().data)
            self.arm_length = int(grp.variables['arm_length'].getValue().data)
        except:
            pass

        self.MAGN_TEMPERATURE = 20
        self.BR_TEMP_COEF = 0
        try:
            grp = ds.groups['magnet']
            self.MAGN_TEMPERATURE = float(
                grp.variables['temperature'].getValue().data)
            self.BR_TEMP_COEF = float(
                grp.variables['br_temp_coef'].getValue().data)
            self.HC_TEMP_COEF = float(
                grp.variables['hc_temp_coef'].getValue().data)
            self.MA_SPEZ_WEIGHT = float(
                grp.variables['spec_mass'].getValue().data)
        except Exception as e:
            #logger.warning("*** MAGNET %s", e)
            pass
        try:
            grp = ds.groups['el_induction']
            (self.curr_loss,
             self.beta_loss,
             self.pos_el_fe_induction,
             self.el_fe_induction_1,
             self.el_fe_induction_2,
             self.eddy_cu_vpot) = [grp.variables[k][:]
                                   for k in ('cur', 'beta', 'position',
                                             'fe_induction_1',
                                             'fe_induction_2',
                                             'eddy_cu_vpot')]
            logger.debug('el_fe_induction %d', len(self.pos_el_fe_induction))
        except KeyError:
            self.pos_el_fe_induction = []
            self.el_fe_induction_1 = []
            self.el_fe_induction_2 = []
            self.eddy_cu_vpot = []

        self.iron_loss_coefficients = []
        try:
            mcgrp = ds.groups['magn_curves']
            lmc = len(mcgrp.variables['ch'])
        except KeyError:
            mcgrp = 0
            lmc = 0

        lcgrp = ds.groups['loss_coeffs']
        ce = 0.0
        ce_freq_exp = 1.5
        ce_ind_exp = 1.5
        ke = 0.0
        khml = 0.65

        lctype = ['magn_curve']*lmc + ['Outside', 'Inside']
        for i in range(lmc+2):
            if lctype[i] == 'magn_curve':
                name = ds.groups['mc'].variables['name'][i].tobytes().decode(
                    'UTF-8').strip()
                base_frequency = float(
                    mcgrp.variables['base_frequency'][i].data)
                base_induction = float(
                    mcgrp.variables['base_induction'][i].data)
                ch = float(mcgrp.variables['ch'][i].data)
                ch_freq_exp = float(mcgrp.variables['ch_exp'][i].data)
                ch_ind_exp = float(mcgrp.variables['ind_exp'][i].data)
                kh = ch/(base_induction**ch_ind_exp *
                         base_frequency**ch_freq_exp)
                cw = float(mcgrp.variables['cw'][i].data)
                cw_freq_exp = float(mcgrp.variables['cw_exp'][i].data)
                cw_ind_exp = float(mcgrp.variables['ind_exp'][i].data)
                spec_weight = float(mcgrp.variables['spec_weight'][i].data)
                fillfactor = float(mcgrp.variables['fillfac'][i].data)
                shapefactor = 1.0
            else:
                j = i - lmc
                name = lctype[i]
                shapefactor = float(
                    lcgrp.variables['shape_factor'][j].data)

                base_frequency = float(lcgrp.variables['freq_0'][j].data)
                base_induction = float(lcgrp.variables['ind_0'][j].data)
                ch = float(lcgrp.variables['ch'][j].data)
                ch_freq_exp = float(lcgrp.variables['fa_hyst'][j].data)
                ch_ind_exp = float(lcgrp.variables['fa_ind'][j].data)
                cw = float(lcgrp.variables['cw'][j].data)
                cw_freq_exp = float(lcgrp.variables['fa_eddy'][j].data)
                cw_ind_exp = float(lcgrp.variables['fa_ind'][j].data)
                spec_weight = float(lcgrp.variables['spec_mass_fe'][j].data)
                fillfactor = float(lcgrp.variables['fill_factor_fe'][j].data)
                shapefactor = float(lcgrp.variables['shape_factor'][j].data)

                self.CU_SPEZ_WEIGHT = float(lcgrp.variables['spec_mass_cu'][j].data)

            kh = ch/(base_induction**ch_ind_exp*base_frequency**ch_freq_exp)

            coeffdict = {
                "Name": name,
                "base_frequency": base_frequency,
                "base_induction": base_induction,
                "ch": ch,
                "ch_freq_exp": ch_freq_exp,
                "ch_ind_exp": ch_ind_exp,
                "kh": kh,
                "khml": khml,
                "cw": cw,
                "cw_freq_exp": cw_freq_exp,
                "cw_ind_exp": cw_ind_exp,
                "ce": ce,
                "ce_freq_exp": ce_freq_exp,
                "ce_ind_exp": ce_ind_exp,
                "ke": ke,
                "spec_weight": spec_weight,
                "fillfactor": fillfactor,
                "shapefactor": shapefactor
            }
            self.iron_loss_coefficients.append(coeffdict)
        ds.close()

def read(filename):
    """
    Read nc file and return NcModel object.

    Arguments:
        filename: name of nc file to be read
    """
    import pathlib
    ncfile = pathlib.Path(filename)
    if ncfile.suffix != '.nc':
        ncfile = ncfile.with_suffix('.nc')
    return isa7.Isa7(Reader(str(ncfile)))


if __name__ == "__main__":
    import sys
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(message)s')
    if len(sys.argv) == 2:
        filename = sys.argv[1]
    else:
        filename = sys.stdin.readline().strip()

    model = read(filename)



# -*- coding: utf-8 -*-
"""
    femagtools.opt
    ~~~~~~~~~~~~~~

    Manage multi-objective optimization with FEMAG



"""
import time
import logging
import pathlib
import femagtools
import femagtools.fsl
import femagtools.moproblem
import femagtools.getset
from .moo.algorithm import Nsga2
from .moo.population import Population
from .femag import set_magnet_properties


logger = logging.getLogger(__name__)


def log_pop(pop, ngen):
    objectives = pop.problem.objective_vars
    decisions = pop.problem.decision_vars
    log = [''.join(['Generation: {}\n'.format(ngen),
                    'rank '] +
                   ["{:10s}".format(o['name'].split('.')[-1])
                    for o in objectives] +
                   [" "] +
                   ["{:12s}".format(d['name'].split('.')[-1])
                    for d in decisions])]

    for i in pop.individuals:
        log.append(''.join(["{} ".format(i.rank)] +
                           ["{:10.2f}".format(f) for f in i.cur_f] +
                           ["   "] +
                           ["{:10.4f}".format(x) for x in i.cur_x]))

    log.append(
        '--------------------------------------------------------------------')
    log.append("  znad: {}\n  zi: {}\n  zw: {}\n  Norm Dist: {}".format(
        pop.compute_nadir(),
        pop.compute_ideal(),
        pop.compute_worst(),
        pop.compute_norm_dist()))
    return log


class Optimizer(object):
    # tasktype='Task'):
    def __init__(self, workdir, magnetizingCurves, magnetMat, condMat=[],
                 result_func=None, templatedirs=[]):
        # self.tasktype = tasktype
        self.templatedirs = templatedirs
        self.result_func = result_func
        self.femag = femagtools.Femag(workdir,
                                      magnetizingCurves=magnetizingCurves,
                                      magnets=magnetMat,
                                      condMat=condMat)

    def _update_population(self, generation, pop, engine):
        self.job.cleanup()

        for k, i in enumerate(pop.individuals):
            task = self.job.add_task(self.result_func)
            pop.problem.prepare(i.cur_x, self.model)
            for mc in self.femag.copy_magnetizing_curves(self.model,
                                                         task.directory):
                task.add_file(mc)
            if 'wdgdef' in self.model.windings:
                self.model.windings['wdgfile'] = self.femag.create_wdg_def(
                    self.model)
            set_magnet_properties(self.model, self.fea, self.femag.magnets)
            task.add_file('femag.fsl',
                          self.builder.create(self.model, self.fea,
                                              self.femag.magnets))
            if 'poc' in self.fea:
                task.add_file(self.fea['pocfilename'],
                              self.fea['poc'].content())
            if 'stateofproblem' in self.fea:
                task.set_stateofproblem(fea['stateofproblem'])
        tstart = time.time()
        ntasks = engine.submit()
        status = engine.join()
        tend = time.time()

        for t, i in zip(self.job.tasks, pop.individuals):
            if t.status == 'C':
                r = t.get_results()
                if isinstance(r, dict) and 'error' in r:
                    logger.warn("Task %s failed: %s", t.id, r['error'])
                else:
                    if isinstance(r, dict):
                        pop.problem.setResult(
                            femagtools.getset.GetterSetter(r))
                    else:
                        pop.problem.setResult(r)

                    i.cur_f = pop.problem.objfun([])
                    i.results = {k: v for k, v in r.items()}
            else:
                logger.warn("Task %s failed with status %s", t.id, t.status)
                i.cur_f = [float('nan')]*pop.problem.f_dim

            i.generation = generation  # for reporting purposes

        pop.update()
        return tend - tstart

    def __call__(self, num_generations, opt, pmMachine,
                 operatingConditions, engine):
        return self.optimize(num_generations, opt, pmMachine,
                             operatingConditions, engine)

    def optimize(self, num_generations, opt, pmMachine,
                 operatingConditions, engine):
        """execute optimization"""
        decision_vars = opt['decision_vars']
        objective_vars = opt['objective_vars']
        population_size = opt['population_size']

        problem = femagtools.moproblem.FemagMoProblem(decision_vars,
                                                      objective_vars)
        self.builder = femagtools.fsl.Builder(self.templatedirs)
        self.model = femagtools.model.MachineModel(pmMachine)
        self.fea = operatingConditions
        self.fea.update(self.model.windings)
        self.fea['lfe'] = self.model.lfe
        self.fea['move_action'] = self.model.move_action
        self.fea['phi_start'] = 0.0
        self.fea['range_phi'] = 720/self.model.get('poles')
        self.pop = Population(problem, population_size)

        algo = Nsga2()

        self.job = engine.create_job(self.femag.workdir)
        # for progress logger
        self.job.num_cur_steps = femagtools.model.FeaModel(
            self.fea).get_num_cur_steps()

        logger.info("Optimize x:%d f:%d generations:%d population size:%d",
                    len(self.pop.problem.decision_vars),
                    len(self.pop.problem.objective_vars),
                    num_generations,
                    self.pop.size())

        results = dict(rank=[], f=[], x=[])
        elapsedTime = 0
        for i in range(num_generations):
            logger.info("Generation %d", i)
            if i > 0:
                newpop = algo.evolve(self.pop)
                deltat = self._update_population(i, newpop, engine)
                self.pop.merge(newpop)
            else:
                deltat = self._update_population(i, self.pop, engine)
            pop_report = '\n'.join(log_pop(self.pop, i) +
                                   ['', '  Elapsed Time: {} s'.format(
                                       int(deltat))])
            repfile = pathlib.Path(
                self.femag.workdir) / f'population-{i}.dat'
            repfile.write_text(pop_report)
            logger.info(pop_report)
            elapsedTime += deltat
        logger.info("TOTAL Elapsed Time: %d s", elapsedTime)
        ft = []
        xt = []
        for i in self.pop.individuals:
            results['rank'].append(i.rank)
            ft.append(i.cur_f)
            xt.append(i.cur_x)
        objective_vars = self.pop.problem.objective_vars
        decision_vars = self.pop.problem.decision_vars
        results['f'] = [[s.get('sign', 1)*y for y in f]
                        for s, f in zip(objective_vars, zip(*ft))]
        results['x'] = list(zip(*xt))
        results['znad'] = [s.get('sign', 1)*v
                           for s, v in zip(objective_vars,
                                           self.pop.compute_nadir())]
        results['zi'] = [s.get('sign', 1)*v
                         for s, v in zip(objective_vars,
                                         self.pop.compute_ideal())]
        results['zw'] = [s.get('sign', 1)*v
                         for s, v in zip(objective_vars,
                                         self.pop.compute_worst())]
        results['dist'] = self.pop.compute_norm_dist()

        def label(d):
            for n in ('desc', 'label', 'name'):
                if n in d:
                    return d
            return '<?>'
        results['objective'] = [label(o) for o in objective_vars]
        results['decision'] = [label(d) for d in decision_vars]
        results['population'] = [i.results
                                 for i in self.pop.individuals if hasattr(i, 'results')]
        return results

    # print("\nChampion: {}\n".format(pop.champion['f']))
        # if flast != None:
        #    print("Fitness Comparison:")
        #    for f1, f2 in zip(pop.champion['f'], flast):
        #        print( "{:10.2f} {:10.2f}      {:10.2f}".format(f1, f2, f1-f2))
        # flast = list(pop.champion['f'])
        # print("")
#    except:
#        print "Unexpected error:", sys.exc_info()

#    print("L2 distance to the best decision vector:")
#    for best_decision in prob.best_x:
#        l2_norm = 0
#        for n in range(0, len(best_decision)):
#            l2_norm +=  (best_decision[n] - isl.population.champion.x[n]) ** 2
#        l2_norm = sqrt(l2_norm)
#        print(l2_norm)

# pf=pop.plot_pareto_fronts()
# savefig('pf0.png')



import pytest
from femagtools import isa7


@pytest.fixture
def isa():
    filename = 'src/tests/data/minimal.ISA7'
    return isa7.read(filename)


def test_read(isa):
    assert len(isa.points) == 39
    assert len(isa.lines) == 39
    assert isa.points[0].x == pytest.approx(0.008339,
                                            abs=1e-5)


def test_objects(isa):
    assert len(isa.points) == 39
    assert len(isa.lines) == 39
    assert len(isa.nodes) == 1729
    assert len(isa.nodechains) == 98
    assert len(isa.elements) == 822
    assert len(isa.superelements) == 3
    assert len(isa.get_subregion('asdf').elements()) == 756


def test_no_such_subregion(isa):
    with pytest.raises(ValueError) as excinfo:
        n = 'foo'
        isa.get_subregion(n)
    assert str(excinfo.value) == (
        'no such subregion "{}" in this model'.format(n))


def test_points(isa):
    p = isa.points[0].x, isa.points[0].y
    p_xy = isa.points[0].xy

    assert p == p_xy == pytest.approx((0.008339, 0.011702), abs=1e-5)


def test_lines(isa):
    ln = isa.lines[0]

    assert type(ln.p1) == isa7.Point
    assert type(ln.p2) == isa7.Point
    assert ln.p1.xy == pytest.approx((0.008339, 0.011702), abs=1e-5)
    assert ln.p2.xy == pytest.approx((0.007388, 0.007190), abs=1e-5)


def test_nodes(isa):
    nd = isa.nodes[-1]

    assert type(nd) == isa7.Node
    assert nd.xy == (nd.x, nd.y) == pytest.approx(
        (0.008787, 0.007736), abs=1e-5)

def test_airgap_center_elements(isa):
    assert len(isa.airgap_center_elements) == 0

def test_nodechains(isa):
    nc = isa.nodechains[0]

    assert type(nc) == isa7.NodeChain
    assert type(nc.nodes[0]) == isa7.Node


def test_nodechains_reverse(isa):
    for nc in isa.nodechains:
        assert nc.reverse().nodes == nc.nodes[::-1]


def test_elements(isa):
    el = isa.elements[0]

    assert type(el) == isa7.Element
    for v in el.vertices:
        assert type(v) == isa7.Node
    assert el.vertices[0].xy == pytest.approx((0.002202, 0.003197),
                                              abs=1e-5)


def test_superelements(isa):
    se = isa.superelements[0]

    assert type(se) == isa7.SuperElement
    for el in se.elements:
        assert type(el) == isa7.Element
    for nc in se.nodechains:
        assert type(nc) == isa7.NodeChain
        assert nc.key == se.nc_keys[se.nodechains.index(nc)]


@pytest.fixture
def disp_stat():
    filename = 'src/tests/data/test_disp_stat.ISA7'
    return isa7.read(filename)


def test_subregions(disp_stat):
    sr = disp_stat.subregions[0]

    assert type(sr) == isa7.SubRegion
    assert sr.name == "Stat"
    for se in sr.superelements:
        assert type(se) == isa7.SuperElement
    for nc in sr.nodechains:
        assert type(nc) == isa7.NodeChain


def test_windings(disp_stat):
    wd = disp_stat.windings[0]

    assert type(wd) == isa7.Winding
    assert wd.name == "Stra"
    for sr in wd.subregions:
        assert type(sr) == isa7.SubRegion
    assert wd.num_turns == 100
